#
# Copyright (C) 2010  Kelvin Lawson (kelvinl@users.sourceforge.net)
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""This module defines the class pykPlayer, which is a base class used
by the modules pykar.py, pycdg.py, and pympg.py.  This collects
together some common interfaces used by these different
implementations for different types of Karaoke files."""

from .pykconstants import *
from .pykmanager import manager
from .pykenv import env
import pygame
import sys
import types
import os

class pykPlayer:
    def __init__(self, song, songDb,
                 errorNotifyCallback = None, doneCallback = None,
                 windowTitle = None):
        """The first parameter, song, may be either a pykdb.SongStruct
        instance, or it may be a filename. """

        if songDb == None:
            from . import pykdb
            songDb = pykdb.globalSongDB
            songDb.LoadSettings(None)
        self.songDb = songDb

        # Set the global command-line options if they have not already
        # been set.
        if manager.options == None:
            parser = self.SetupOptions()
            (manager.options, args) = parser.parse_args()
            manager.ApplyOptions(self.songDb)

            if song is None:
                if (len(args) != 1):
                    parser.print_help()
                    sys.exit(2)
                song = args[0]

        # Unfortunately, we can't capture sound when dumping.  There
        # are two reasons for this.  (1) pymedia doesn't currently
        # support multiplexing audio with a video stream, so when
        # you're dumping an mpeg file, it has to be video-only.  (2)
        # pygame doesn't provide a way for us to programmatically
        # convert a midi file to sound samples anyway--all you can do
        # with a midi file is route it through the speakers.

        # So, for these reasons, we always just disable sound when
        # dumping images or movies.
        if manager.options.dump:
            manager.options.nomusic = True

        if isinstance(song, (str,)):
            # We were given a filename.  Convert it to a SongStruct.
            song = self.songDb.makeSongStruct(song)
        
        # Store the parameters
        self.Song = song
        self.WindowTitle = windowTitle

        # And look up the actual files corresponding to this SongStruct.
        self.SongDatas = song.GetSongDatas()
        if windowTitle is None:
            self.WindowTitle = song.DisplayFilename
            
        # Caller can register a callback by which we
        # print out error information, use stdout if none registered
        if errorNotifyCallback:
            self.ErrorNotifyCallback = errorNotifyCallback
        else:
            self.ErrorNotifyCallback = self.__defaultErrorPrint
    
        # Caller can register a callback by which we
        # let them know when the song is finished
        if doneCallback:
            self.SongFinishedCallback = doneCallback
        else:
            self.SongFinishedCallback = None

        self.State = STATE_INIT
        self.InternalOffsetTime = 0

        # These values are used to keep track of the current position
        # through the song based on pygame's get_ticks() interface.
        # It's used only when get_pos() cannot be used or is
        # unreliable for some reason.
        self.PlayTime = 0
        self.PlayStartTime = 0
        self.PlayFrame = 0

        # self.PlayStartTime is valid while State == STATE_PLAYING; it
        # indicates the get_ticks() value at which the song started
        # (adjusted for any pause intervals that occurred during
        # play).  self.PlayTime is valid while State != STATE_PLAYING;
        # it indicates the total number of ticks (milliseconds) that
        # have elapsed in the song so far.

        # self.PlayFrame starts at 0 and increments once for each
        # frame.  It's not very meaningful, except in STATE_CAPTURING
        # mode.

        # Keep track of the set of modifier buttons that are held
        # down.  This is currently used only for the GP2X interface.
        self.ShoulderLHeld = False
        self.ShoulderRHeld = False

        # Set this true if the player can zoom font sizes.
        self.SupportsFontZoom = False

    # The following methods are part of the public API and intended to
    # be exported from this class.

    def Validate(self):
        """ Returns True if the karaoke file appears to be playable
        and contains lyrics, or False otherwise. """

        return self.doValidate()

    def Play(self):
        self.doPlay()

        if manager.options.dump:
            self.setupDump()
        else:
            self.PlayStartTime = pygame.time.get_ticks()
            self.State = STATE_PLAYING

    # Pause the song - Use Pause() again to unpause
    def Pause(self):
        if self.State == STATE_PLAYING:
            self.doPause()
            self.PlayTime = pygame.time.get_ticks() - self.PlayStartTime
            self.State = STATE_PAUSED
        elif self.State == STATE_PAUSED:
            self.doUnpause()
            self.PlayStartTime = pygame.time.get_ticks() - self.PlayTime
            self.State = STATE_PLAYING

    # Close the whole thing down
    def Close(self):
        self.State = STATE_CLOSING

    # you must call Play() to restart. Blocks until pygame is initialised
    def Rewind(self):
        self.doRewind()
        self.PlayTime = 0
        self.PlayStartTime = 0
        self.PlayFrame = 0
        self.State = STATE_NOT_PLAYING

    # Stop the song and go back to the start. As you would
    # expect Stop to do on a CD player. Play() restarts from
    # the beginning
    def Stop(self):
        self.Rewind()
            
    # Get the song length (in seconds)
    def GetLength(self):
        ErrorString = "GetLength() not supported"
        self.ErrorNotifyCallback (ErrorString)
        return None

    # Get the current time (in milliseconds).
    def GetPos(self):
        if self.State == STATE_PLAYING:
            return pygame.time.get_ticks() - self.PlayStartTime
        else:
            return self.PlayTime

    def SetupOptions(self, usage = None):
        """ Initialise and return optparse OptionParser object,
        suitable for parsing the command line options to this
        application. """

        if usage == None:
            usage = "%prog [options] <Karaoke file>"

        return manager.SetupOptions(usage, self.songDb)

    # Below methods are internal.

    def setupDump(self):
        # Capture the output as a sequence of numbered frame images.
        self.PlayTime = 0
        self.PlayStartTime = 0
        self.PlayFrame = 0
        self.State = STATE_CAPTURING

        self.dumpFrameRate = manager.options.dump_fps
        assert self.dumpFrameRate

        filename = manager.options.dump
        base, ext = os.path.splitext(filename)
        ext_lower = ext.lower()

        self.dumpEncoder = None
        if ext_lower == '.mpg':
            # Use pymedia to convert frames to an mpeg2 stream
            # on-the-fly.
            import pymedia
            import pymedia.video.vcodec as vcodec

            self.dumpFile = open(filename, 'wb')
            frameRate = int(self.dumpFrameRate * 100 + 0.5)
            self.dumpFrameRate = float(frameRate) / 100.0
            
            params= { \
              'type': 0,
              'gop_size': 12,
              'frame_rate_base': 125,
              'max_b_frames': 0,
              'height': manager.options.size_y,
              'width': manager.options.size_x,
              'frame_rate': frameRate,
              'deinterlace': 0,
              'bitrate': 9800000,
              'id': vcodec.getCodecID('mpeg2video')
            }
            self.dumpEncoder = vcodec.Encoder( params )
            return
            
        # Don't dump a video file; dump a sequence of frames instead.
        self.dumpPPM = (ext_lower == '.ppm' or ext_lower == '.pnm')
        self.dumpAppend = False
        
        # Convert the filename to a pattern.
        if '#' in filename:
            hash = filename.index('#')
            end = hash
            while end < len(filename) and filename[end] == '#':
                end += 1
            count = end - hash
            filename = filename[:hash] + '%0' + str(count) + 'd' + filename[end:]
        else:
            # There's no hash in the filename.
            if self.dumpPPM:
                # We can dump a series of frames all to the same file,
                # if we're dumping ppm frames.  Mjpegtools likes this.
                self.dumpAppend = True
                try:
                    os.remove(filename)
                except OSError:
                    pass
            else:
                # Implicitly append a frame number.
                filename = base + '%04d' + ext
            
        self.dumpFilename = filename

    def doFrameDump(self):
        if self.dumpEncoder:
            import pymedia.video.vcodec as vcodec

            ss = pygame.image.tostring(manager.surface, "RGB")
            bmpFrame = vcodec.VFrame(
                vcodec.formats.PIX_FMT_RGB24,
                manager.surface.get_size(), (ss,None,None))
            yuvFrame = bmpFrame.convert(vcodec.formats.PIX_FMT_YUV420P)
            d = self.dumpEncoder.encode(yuvFrame)
            self.dumpFile.write(d.data)
            return

        if self.dumpAppend:
            filename = self.dumpFilename
        else:
            filename = self.dumpFilename % self.PlayFrame
            print(filename)

        if self.dumpPPM:
            # Dump a PPM file.  We do PPM by hand since pygame
            # doesn't support it directly, but it's so easy and
            # useful.
            
            w, h = manager.surface.get_size()
            if self.dumpAppend:
                f = open(filename, 'ab')
            else:
                f = open(filename, 'wb')
            f.write('P6\n%s %s 255\n' % (w, h))
            f.write(pygame.image.tostring(manager.surface, 'RGB'))

        else:
            # Ask pygame to dump the file.  We trust that pygame knows
            # how to store an image in the requested format.
            pygame.image.save(manager.surface, filename)
            

    def doValidate(self):
        return True
    
    def doPlay(self):
        pass

    def doPause(self): 
        pass

    def doUnpause(self):
        pass

    def doRewind(self):
        pass

    def doStuff(self):
        # Override this in a derived class to do some useful per-frame
        # activity.
        # Common handling code for a close request or if the
        # pygame window was quit
        if self.State == STATE_CLOSING:
            if manager.display:
                manager.display.fill((0,0,0))
                pygame.display.flip()
            self.shutdown()

        elif self.State == STATE_CAPTURING:
            # We are capturing a video file.
            self.doFrameDump()

            # Set the frame time for the next frame.
            self.PlayTime = 1000.0 * self.PlayFrame / self.dumpFrameRate
            
        self.PlayFrame += 1
        

    def doResize(self, newSize):
        # This will be called internally whenever the window is
        # resized for any reason, either due to an application resize
        # request being processed, or due to the user dragging the
        # window handles.
        pass

    def doResizeBegin(self):
        # This will be called internally before the screen is resized
        # by pykmanager and doResize() is called. Not all players need
        # to do anything here.
        pass

    def doResizeEnd(self):
        # This will be called internally after the screen is resized
        # by pykmanager and doResize() is called. Not all players need
        # to do anything here.
        pass

    def handleEvent(self, event):
        if event.type == pygame.USEREVENT:
            self.Close()
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
                self.Close()

            elif event.key == pygame.K_PAUSE or event.key == pygame.K_p:
                self.Pause()

            elif event.key == pygame.K_BACKSPACE or event.key == pygame.K_DELETE:
                self.Rewind()
                self.Play()

            # Use control-left/right arrow to offset the current
            # graphics time by 1/4 sec.  Use control-down arrow to
            # restore them to sync.
            elif self.State == STATE_PLAYING and event.key == pygame.K_RIGHT and event.mod & (pygame.KMOD_LCTRL | pygame.KMOD_RCTRL):
                manager.settings.SyncDelayMs += 250
                print("sync %s" % manager.settings.SyncDelayMs)
            elif self.State == STATE_PLAYING and event.key == pygame.K_LEFT and event.mod & (pygame.KMOD_LCTRL | pygame.KMOD_RCTRL):
                manager.settings.SyncDelayMs -= 250
                print("sync %s" % manager.settings.SyncDelayMs)
            elif self.State == STATE_PLAYING and event.key == pygame.K_DOWN and event.mod & (pygame.KMOD_LCTRL | pygame.KMOD_RCTRL):
                manager.settings.SyncDelayMs = 0
                print("sync %s" % manager.settings.SyncDelayMs)

            if self.SupportsFontZoom:
                if event.key == pygame.K_PLUS or event.key == pygame.K_EQUALS or \
                   event.key == pygame.K_KP_PLUS:
                    manager.ZoomFont(1.0/0.9)
                elif event.key == pygame.K_MINUS or event.key == pygame.K_UNDERSCORE or \
                   event.key == pygame.K_KP_MINUS:
                    manager.ZoomFont(0.9)

        elif event.type == pygame.QUIT:
            self.Close()
            
        elif env == ENV_GP2X and event.type == pygame.JOYBUTTONDOWN:
            if event.button == GP2X_BUTTON_SELECT:
                self.Close()
            elif event.button == GP2X_BUTTON_START:
                self.Pause()
            elif event.button == GP2X_BUTTON_L:
                self.ShoulderLHeld = True
            elif event.button == GP2X_BUTTON_R:
                self.ShoulderRHeld = True

            if self.SupportsFontZoom:
                if event.button == GP2X_BUTTON_RIGHT and self.ShoulderLHeld:
                    manager.ZoomFont(1.0/0.9)
                elif event.button == GP2X_BUTTON_LEFT and self.ShoulderLHeld:
                    manager.ZoomFont(0.9)
            
        elif env == ENV_GP2X and event.type == pygame.JOYBUTTONUP:
            if event.button == GP2X_BUTTON_L:
                self.ShoulderLHeld = False
            elif event.button == GP2X_BUTTON_R:
                self.ShoulderRHeld = False

    def shutdown(self):
        # This will be called by the pykManager to shut down the thing
        # immediately.

        # If the caller gave us a callback, let them know we're finished
        if self.State != STATE_CLOSED:
            self.State = STATE_CLOSED
            if self.SongFinishedCallback != None:
                self.SongFinishedCallback()


    def __defaultErrorPrint(self, ErrorString):
        print (ErrorString)


    def findPygameFont(self, fontData, fontSize):
        """ Returns a pygame.Font selected by this data. """
        if not fontData.size:
            # The font names a specific filename.
            filename = fontData.name
            if os.path.sep not in filename:
                filename = os.path.join(manager.FontPath, filename)
            return pygame.font.Font(filename, fontSize)

        # The font names a system font.
        pointSize = int(fontData.size * fontSize / 10.0 + 0.5)
        return pygame.font.SysFont(
            fontData.name, pointSize, bold = fontData.bold,
            italic = fontData.italic)



#******************************************************************************
#****                                                                      ****
#**** Copyright (C) 2010  Kelvin Lawson (kelvinl@users.sourceforge.net)    ****
#**** Copyright (C) 2010  PyKaraoke Development Team                       ****
#****                                                                      ****
#**** This library is free software; you can redistribute it and/or        ****
#**** modify it under the terms of the GNU Lesser General Public           ****
#**** License as published by the Free Software Foundation; either         ****
#**** version 2.1 of the License, or (at your option) any later version.   ****
#****                                                                      ****
#**** This library is distributed in the hope that it will be useful,      ****
#**** but WITHOUT ANY WARRANTY; without even the implied warranty of       ****
#**** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    ****
#**** Lesser General Public License for more details.                      ****
#****                                                                      ****
#**** You should have received a copy of the GNU Lesser General Public     ****
#**** License along with this library; if not, write to the                ****
#**** Free Software Foundation, Inc.                                       ****
#**** 59 Temple Place, Suite 330                                           ****
#**** Boston, MA  02111-1307  USA                                          ****
#******************************************************************************

from .pykconstants import *
from .pykenv import env
from . import pykversion
import pygame
import os
import sys

# Python 2.3 and newer ship with optparse; older Python releases need "Optik"
# installed (optik.sourceforge.net)
try:
    import optparse
except:
    import Optik as optparse

if env == ENV_GP2X:
    import _cpuctrl as cpuctrl

class pykManager:

    """ There is only one instance of this class in existence during
    program execution, and it is never destructed until program
    termination.  This class manages the pygame interface, keeping
    interfaces open or closed as necessary; it also provides callbacks
    into handling pygame events. """

    def __init__(self):
        self.initialized = False
        self.player = None
        self.options = None
        self.display = None
        self.surface = None
        self.audioProps = None

        self.displaySize = None
        self.displayFlags = 0
        self.displayDepth = 0
        self.cpuSpeed = None

        # Find the correct font path. If fully installed on Linux this
        # will be sys.prefix/share/pykaraoke/fonts. Otherwise look for
        # it in the current directory.
        if (os.path.isfile("fonts/DejaVuSans.ttf")):
            self.FontPath = "fonts"
            self.IconPath = "icons"
        else:
            self.FontPath = os.path.join(sys.prefix, "share/pykaraoke/fonts")
            self.IconPath = os.path.join(sys.prefix, "share/pykaraoke/icons")

        if env == ENV_GP2X:
            speed = cpuctrl.get_FCLK()
            print("Initial CPU speed is %s" % (speed))
            x, y, tvout = cpuctrl.get_screen_info()
            print("Initial screen size is %s, %s" % (x, y))
            if tvout:
                print("TV-Out mode is enabled.")

        # This factor may be changed by the user to make text bigger
        # or smaller on those players that support it.
        self.fontScale = None

    def setCpuSpeed(self, activityName):
        """ Sets the CPU speed appropriately according to what the
        current activity is.  At the moment, this is used only for the
        GP2X. """

        if self.cpuSpeed == activityName:
            # No change.
            return
        self.cpuSpeed = activityName

        # The activityName directly hooks into a CPU speed indicated
        # in the user settings.

        attr = 'CPUSpeed_%s' % (activityName)
        speed = getattr(self.settings, attr, None)
        if speed is not None:
            self.OpenCPUControl()
            if env == ENV_GP2X:
                cpuctrl.set_FCLK(speed)
                pass

    def VolumeUp(self):
        try:
            volume = pygame.mixer.music.get_volume()
        except pygame.error:
            print("Failed to raise music volume!")
            return
        volume = min(volume + 0.1, 1.0)

        pygame.mixer.music.set_volume(volume)

    def VolumeDown(self):
        try:
            volume = pygame.mixer.music.get_volume()
        except pygame.error:
            print("Failed to lower music volume!")
            return
        volume = max(volume - 0.1, 0.0)

        pygame.mixer.music.set_volume(volume)

    def GetVolume(self):
        """ Gives the current volume level. """
        if 'music' in vars():
            return pygame.mixer.music.get_volume()
        else:
            return 0.50 # 75% is the industry recommended maximum value

    def SetVolume(self, volume):
        """ Sets the volume of the music playback. """
        volume = min(volume, 1.0)
        volume = max(volume, 0.0)
        pygame.mixer.music.set_volume(volume)

    def GetFontScale(self):
        """ Returns the current font scale. """
        if self.fontScale == None:
            self.fontScale = self.options.font_scale
        return self.fontScale

    def ZoomFont(self, factor):
        """ Zooms the font scale by the indicated factor.  This is
        treated like a resize event, even though the window is not
        changing size; player.doResize() will be called. """
        self.GetFontScale()
        self.fontScale *= factor
        if self.player:
            self.player.doResize(self.displaySize)

    def InitPlayer(self, player):

        """ A pykPlayer will call this when it constructs.  This
        registers the player with the pykManager, so that it will get
        callbacks and control of the display.  This call also ensures
        that pygame has been initialized. """

        if self.player:
            self.player.shutdown()
            self.player = None

        # Ensure we have been initialized.
        if not self.initialized:
            self.pygame_init()

        self.player = player
        self.player.State = STATE_NOT_PLAYING

        if self.display != None and self.displayTitle == None:
            try:
                pygame.display.set_caption(player.WindowTitle)
            except UnicodeError:
                pygame.display.set_caption(player.WindowTitle.encode('UTF-8', 'replace'))


    def OpenDisplay(self, displaySize = None, flags = None, depth = None):
        """ Use this method to open a pygame display or set the
        display to a specific mode. """

        self.getDisplayDefaults()

        if displaySize == None:
            displaySize = self.displaySize
        if flags == None:
            flags = self.displayFlags
        if depth == None:
            depth = self.displayDepth

        if self.options.dump:
            # We're just capturing frames offscreen.  In that case,
            # just open an offscreen buffer as the "display".
            self.display = None
            self.surface = pygame.Surface(self.displaySize)
            self.mouseVisible = False
            self.displaySize = self.surface.get_size()
            self.displayFlags = self.surface.get_flags()
            self.displayDepth = self.surface.get_bitsize()
        else:
            # Open the onscreen display normally.
            pygame.display.init()

            self.mouseVisible = not (env == ENV_GP2X or self.options.hide_mouse or (self.displayFlags & pygame.FULLSCREEN))
            pygame.mouse.set_visible(self.mouseVisible)

            if self.displayTitle != None:
                pygame.display.set_caption(self.displayTitle)
            elif self.player != None:
                try:
                    pygame.display.set_caption(self.player.WindowTitle)
                except UnicodeError:
                    pygame.display.set_caption(self.player.WindowTitle.encode('UTF-8', 'replace'))

            if self.display == None or \
               (self.displaySize, self.displayFlags, self.displayDepth) != (displaySize, flags, depth):
                ## displaySize = pygame.display.list_modes()[0]
                ## flags = pygame.FULLSCREEN
                displaySize = (1920,1080)
                self.display = pygame.display.set_mode(displaySize, flags, depth)
                self.displaySize = self.display.get_size()
                self.displayFlags = flags
                self.displayDepth = depth

            self.surface = self.display

        self.displayTime = pygame.time.get_ticks()

    def Flip(self):
        """ Call this method to make the displayed frame visible. """
        if self.display:
            pygame.display.flip()

    def CloseDisplay(self):
        """ Use this method to close the pygame window if it has been
        opened. """

        if self.display:
            pygame.display.quit()
            pygame.display.init()
            self.display = None

        self.surface = None

    def OpenAudio(self, frequency = None, size = None, channels = None):
        """ Use this method to initialize or change the audio
        parameters."""

        # We shouldn't mess with the CPU control while the audio is
        # open.
        self.CloseCPUControl()

        if frequency == None:
            frequency = self.settings.SampleRate

        if size == None:
            size = -16

        if channels == None:
            channels = self.settings.NumChannels

        bufferMs = self.settings.BufferMs

        # Compute the number of samples that would fill the indicated
        # buffer time.
        bufferSamples = bufferMs * (frequency * channels) / 1000

        # This needs to be a power of 2, so find the first power of 2
        # larger, up to 2^15.
        p = 1
        while p < bufferSamples and p < 32768:
            p <<= 1
        # Now choose the power of 2 closest.

        if (abs(bufferSamples - (p >> 1)) < abs(p - bufferSamples)):
            bufferSamples = p >> 1
        else:
            bufferSamples = p

        audioProps = (frequency, size, channels, bufferSamples)
        if audioProps != self.audioProps:
            # If the audio properties have changed, we have to shut
            # down and re-start the audio subsystem.
            pygame.mixer.quit()
            pygame.mixer.init(*audioProps)
            self.audioProps = audioProps

    def CloseAudio(self):
        pygame.mixer.quit()
        self.audioProps = None

    def OpenCPUControl(self):
        self.CloseAudio()
        if env == ENV_GP2X:
            cpuctrl.init()

    def CloseCPUControl(self):
        if env == ENV_GP2X:
            cpuctrl.shutdown()

    def GetAudioBufferMS(self):
        """ Returns the number of milliseconds it will take to
        completely empty a full audio buffer with the current
        settings. """
        if self.audioProps:
            frequency, size, channels, bufferSamples = self.audioProps
            return bufferSamples * 1000 / (frequency * channels)
        return 0

    def Quit(self):
        if self.player:
            self.player.shutdown()
            self.player = None

        if not self.initialized:
            return
        self.initialized = False

        pygame.quit()

    def __errorCallback(self, message):
        self.songValid = False
        print(message)
    def __doneCallback(self):
        pass

    def ValidateDatabase(self, songDb):
        """ Validates all of the songs in the database, to ensure they
        are playable and contain lyrics. """

        self.CloseDisplay()
        invalidFile = open('invalid.txt', 'w')

        songDb.SelectSort('filename')
        for song in songDb.SongList[:1074]:
            self.songValid = True
            player = song.MakePlayer(songDb, self.__errorCallback, self.__doneCallback)
            if not player:
                self.songValid = False
            else:
                if not player.Validate():
                    self.songValid = False

            if self.songValid:
                print('%s ok' % (song.DisplayFilename))
            else:
                print('%s invalid' % (song.DisplayFilename))
                print('%s\t%s' % (song.Filepath, song.ZipStoredName), file=invalidFile)
                invalidFile.flush()

    def Poll(self):
        """ Your application must call this method from time to
        time--ideally, within a hundred milliseconds or so--to perform
        the next quantum of activity. Alternatively, if the
        application does not require any cycles, you may just call
        WaitForPlayer() instead. """

        if not self.initialized:
            self.pygame_init()

        self.handleEvents()

        if self.player:
            if self.player.State == STATE_CLOSED:
                self.player = None
            else:
                self.player.doStuff()

        # Wait a bit to save on wasteful CPU usage.
        pygame.time.wait(1)

    def WaitForPlayer(self):
        """ The interface may choose to call this method in lieu of
        repeatedly calling Poll().  It will block until the currently
        active player has finished, and then return. """

        while self.player and self.player.State != STATE_CLOSED:
            self.Poll()

    def SetupOptions(self, usage, songDb):
        """ Initialise and return optparse OptionParser object,
        suitable for parsing the command line options to this
        application.  This version of this method returns the options
        that are likely to be useful for any karaoke application. """

        version = "%prog " + pykversion.PYKARAOKE_VERSION_STRING

        settings = songDb.Settings

        parser = optparse.OptionParser(usage = usage, version = version,
                                       conflict_handler = "resolve")

        if env != ENV_OSX and env != ENV_GP2X:
            pos_x = None
            pos_y = None
            if settings.PlayerPosition:
                pos_x, pos_y = settings.PlayerPosition
            parser.add_option('-x', '--window-x', dest = 'pos_x', type = 'int', metavar='X',
                              help = 'position song window X pixels from the left edge of the screen', default = pos_x)
            parser.add_option('-y', '--window-y', dest = 'pos_y', type = 'int', metavar='Y',
                              help = 'position song window Y pixels from the top edge of the screen', default = pos_y)

        if env != ENV_GP2X:
            parser.add_option('-w', '--width', dest = 'size_x', type = 'int', metavar='X',
                              help = 'draw song window X pixels wide', default = settings.PlayerSize[0])
            parser.add_option('-h', '--height', dest = 'size_y', type = 'int', metavar='Y',
                              help = 'draw song window Y pixels high', default = settings.PlayerSize[1])
            parser.add_option('-t', '--title', dest = 'title', type = 'string', metavar='TITLE',
                              help = 'set song window title to TITLE', default = None)
            parser.add_option('-f', '--fullscreen', dest = 'fullscreen', action = 'store_true',
                              help = 'make song window fullscreen', default = settings.FullScreen)
            parser.add_option('', '--hide-mouse', dest = 'hide_mouse', action = 'store_true',
                              help = 'hide the mouse pointer', default = False)

        parser.add_option('-s', '--fps', dest = 'fps', metavar='N', type = 'int',
                          help = 'restrict visual updates to N frames per second',
                          default = 30)
        parser.add_option('-r', '--sample-rate', dest = 'sample_rate', type = 'int',
                          help = 'specify the audio sample rate.  Ideally, this should match the recording.  For MIDI files, higher is better but consumes more CPU.',
                          default = settings.SampleRate)
        parser.add_option('', '--num-channels', dest = 'num_channels', type = 'int',
                          help = 'specify the number of audio channels: 1 for mono, 2 for stereo.',
                          default = settings.NumChannels)
        parser.add_option('', '--font-scale', metavar='SCALE', dest = 'font_scale', type = 'float',
                          help = 'specify the font scale factor; small numbers (between 0 and 1) make text smaller so more fits on the screen, while large numbers (greater than 1) make text larger so less fits on the screen.',
                          default = 1)

        parser.add_option('', '--zoom', metavar='MODE', dest = 'zoom_mode', type = 'choice',
                          choices = settings.Zoom,
                          help = 'specify the way in which graphics are scaled to fit the window.  The choices are %s.' % (', '.join(['"%s"' % z for z in settings.Zoom])),
                          default = settings.CdgZoom)

        parser.add_option('', '--buffer', dest = 'buffer', metavar = 'MS', type = 'int',
                          help = 'buffer audio by the indicated number of milliseconds',
                          default = settings.BufferMs)
        parser.add_option('-n', '--nomusic', dest = 'nomusic', action = 'store_true',
                          help = 'disable music playback, just display graphics', default = False)

        parser.add_option('', '--dump', dest = 'dump',
                          help = 'dump output as a sequence of frame images, for converting to video',
                          default = '')
        parser.add_option('', '--dump-fps', dest = 'dump_fps', type = 'float',
                          help = 'specify the number of frames per second of the sequence output by --dump',
                          default = 29.97)

        parser.add_option('', '--validate', dest = 'validate', action = 'store_true',
                          help = 'validate that all songs contain lyrics and are playable')

        return parser

    def ApplyOptions(self, songDb):
        """ Copies the user-specified command-line options in
        self.options to the settings in songDb.Settings. """

        self.settings = songDb.Settings

        self.settings.CdgZoom = self.options.zoom_mode

        if hasattr(self.options, 'fullscreen'):
            self.settings.FullScreen = self.options.fullscreen
            self.settings.PlayerSize = (self.options.size_x, self.options.size_y)
        if hasattr(self.options, 'pos_x') and \
           self.options.pos_x != None and self.options.pos_y != None:
            self.settings.PlayerPosition = (self.options.pos_x, self.options.pos_y)

        self.settings.NumChannels = self.options.num_channels
        self.settings.SampleRate = self.options.sample_rate
        self.settings.BufferMs = self.options.buffer

    def WordWrapText(self, text, font, maxWidth):
        """Folds the line (or lines) of text into as many lines as
        necessary to fit within the indicated width (when rendered by
        the given font), word-wrapping at spaces.  Returns a list of
        strings, one string for each separate line."""


        lines = []

        for line in text.split('\n'):
            fold = self.FindFoldPoint(line, font, maxWidth)
            while line:
                lines.append(line[:fold])
                line = line[fold:]
                fold = self.FindFoldPoint(line, font, maxWidth)

        return lines

    def FindFoldPoint(self, line, font, maxWidth):
        """Returns the index of the character within line which should
        begin the next line: the first non-space before maxWidth."""

        if maxWidth <= 0 or line == '':
            return len(line)

        fold = len(line.rstrip())
        width, height = font.size(line[:fold])
        while fold > 0 and width > maxWidth:
            sp = line[:fold].rfind(' ')
            if sp == -1:
                fold -= 1
            else:
                fold = sp
            width, height = font.size(line[:fold])

        while fold < len(line) and line[fold] == ' ':
            fold += 1

        if fold == 0:
            # Couldn't even get one character in.  Put it in anyway.
            fold = 1

        if line[:fold].strip() == '':
            # Oops, nothing but whitespace in front of the fold.  Try
            # again without the whitespace.
            ws = line[:fold]
            line = line[fold:]
            wsWidth, height = font.size(ws)
            return self.FindFoldPoint(line, font, maxWidth - wsWidth) + len(ws)

        return fold


    # The remaining methods are internal.

    def handleEvents(self):
        """ Handles the events returned from pygame. """

        if self.display:
            # check for Pygame events
            for event in pygame.event.get():
                self.handleEvent(event)


    def handleEvent(self, event):
        # Only handle resize events 250ms after opening the
        # window. This is to handle the bizarre problem of SDL making
        # the window small automatically if you set
        # SDL_VIDEO_WINDOW_POS and move the mouse around while the
        # window is opening. Give it some time to settle.
        player = self.player
        if event.type == pygame.VIDEORESIZE and pygame.time.get_ticks() - self.displayTime > 250:

            # Tell the player we are about to resize. This is required
            # for pympg.
            if player:
                player.doResizeBegin()

            # Do the resize
            self.displaySize = event.size
            self.settings.PlayerSize = tuple(self.displaySize)
            pygame.display.set_mode(event.size, self.displayFlags, self.displayDepth)
            # Call any player-specific resize
            if player:
                player.doResize(event.size)

            # Tell the player we have finished resizing
            if player:
                player.doResizeEnd()

        elif env == ENV_GP2X and event.type == pygame.JOYBUTTONDOWN:
            if event.button == GP2X_BUTTON_VOLUP:
                self.VolumeUp()
            elif event.button == GP2X_BUTTON_VOLDOWN:
                self.VolumeDown()

        if player:
            player.handleEvent(event)

    def pygame_init(self):
        """ This method is called only once, the first time an
        application requests a pygame window. """
        os.environ['SDL_VIDEO_CENTERED'] = '1'

        pygame.init()

        if env == ENV_GP2X:
            num_joysticks = pygame.joystick.get_count()
            if num_joysticks > 0:
                stick = pygame.joystick.Joystick(0)
                stick.init() # now we will receive events for the GP2x joystick and buttons

        self.initialized = True

    def getDisplayDefaults(self):
        if env == ENV_GP2X:
            # The GP2x has no control over its window size or
            # placement.  You'll get fullscreen and like it.

            # Unfortunately, it appears that pygame--or maybe our SDL,
            # even though we'd compiled with paeryn's HW SDL--doesn't
            # allow us to open a TV-size window, so we have to settle
            # for the standard (320, 240) size and let the hardware
            # zooming scale it for TV out.
            self.displaySize = (320, 240)
            self.displayFlags = pygame.HWSURFACE | pygame.FULLSCREEN
            self.displayDepth = 0
            self.displayTitle = None
            self.mouseVisible = False
            return

        # Fix the position at top-left of window. Note when doing
        # this, if the mouse was moving around as the window opened,
        # it made the window tiny.  Have stopped doing anything for
        # resize events until 1sec into the song to work around
        # this. Note there appears to be no way to find out the
        # current window position, in order to bring up the next
        # window in the same place. Things seem to be different in
        # development versions of pygame-1.7 - it appears to remember
        # the position, and it is the only version for which fixing
        # the position works on MS Windows.

        # Don't set the environment variable on OSX.
        if env != ENV_OSX:
            if self.settings.PlayerPosition:
                x, y = self.settings.PlayerPosition
                os.environ['SDL_VIDEO_WINDOW_POS'] = "%s,%s" % (x, y)

        w, h = self.settings.PlayerSize
        self.displaySize = (w, h)

        self.displayFlags = pygame.RESIZABLE
        if self.settings.DoubleBuf:
            self.displayFlags |= pygame.DOUBLEBUF
        if self.settings.HardwareSurface:
            self.displayFlags |= pygame.HWSURFACE
        
        if self.settings.NoFrame:
            self.displayFlags |= pygame.NOFRAME
        if self.settings.FullScreen:
            self.displayFlags |= pygame.FULLSCREEN

        self.displayDepth = 0
        self.displayTitle = self.options.title

        self.mouseVisible = not (env == ENV_GP2X or self.options.hide_mouse or (self.displayFlags & pygame.FULLSCREEN))


# Now instantiate a global pykManager object.
manager = pykManager()



#!/usr/bin/env python 

# pympg - MPEG Karaoke Player
#
# Copyright (C) 2010  Kelvin Lawson (kelvinl@users.sourceforge.net)
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from .pykconstants import *
from .pykplayer import pykPlayer
from .pykenv import env
from .pykmanager import manager
import pygame, sys, os, string, subprocess
import threading

# OVERVIEW
#
# pympg is an MPEG player built using python. It was written for the
# PyKaraoke project but is in fact a general purpose MPEG player that
# could be used in other python projects requiring an MPEG player.
#
# The player uses the pygame library (www.pygame.org), and can therefore
# run on any operating system that runs pygame (currently Linux, Windows
# and OSX).
#
# You can use this file as a standalone player, or together with
# PyKaraoke. PyKaraoke provides a graphical user interface, playlists,
# searchable song database etc.
#
# For those writing a media player or similar project who would like
# MPG support, this module has been designed to be easily incorporated
# into such projects and is released under the LGPL.


# REQUIREMENTS
#
# pympg requires the following to be installed on your system:
# . Python (www.python.org)
# . Pygame (www.pygame.org)


# USAGE INSTRUCTIONS
#
# To start the player, pass the MPEG filename/path on the command line:
#       python pympg.py /songs/theboxer.mpg
#
# You can also incorporate a MPG player in your own projects by
# importing this module. The class mpgPlayer is exported by the
# module. You can import and start it as follows:
#   import pympg
#   player = pympg.mpgPlayer("/songs/theboxer.mpg")
#   player.Play()
# If you do this, you must also arrange to call pympg.manager.Poll()
# from time to time, at least every 100 milliseconds or so, to allow
# the player to do its work.
#
# The class also exports Close(), Pause(), Rewind(), GetPos().
#
# There are two optional parameters to the initialiser, errorNotifyCallback
# and doneCallback:
#
# errorNotifyCallback, if provided, will be used to print out any error
# messages (e.g. song file not found). This allows the module to fit 
# together well with GUI playlist managers by utilising the same GUI's
# error popup window mechanism (or similar). If no callback is provided,
# errors are printed to stdout. errorNotifyCallback should take one 
# parameter, the error string, e.g.:
#   def errorPopup (ErrorString):
#       msgBox (ErrorString)
#
# doneCallback can be used to register a callback so that the player
# calls you back when the song is finished playing. The callback should
# take no parameters, e.g.:
#   def songFinishedCallback():
#       msgBox ("Song is finished")
#
# To register callbacks, pass the functions in to the initialiser:
#   mpgPlayer ("/songs/theboxer.mpg", errorPopup, songFinishedCallback)
# These parameters are optional and default to None.
#
# If the initialiser fails (e.g. the song file is not present), __init__
# raises an exception.


# IMPLEMENTATION DETAILS
#
# pympg is implemented as a handful of python modules. Pygame provides
# all of the MPEG decoding and display capabilities, and can play an
# MPEG file with just a few lines of code. Hence this module is rather
# small. What it provides on top of the basic pygame features, is a
# player-like class interface with Play, Pause, Rewind etc. It also
# implements a resizable player window.  And, of course, it integrates
# nicely with pykaraoke.py and pykaraoke_mini.py.
#
# Previous implementations ran the player within a thread; this is no
# longer the case.  Instead, it is the caller's responsibility to call
# pycdg.manager.Poll() every once in a while to ensure that the player
# gets enough CPU time to do its work.  Ideally, this should be at
# least every 100 milliseconds or so to guarantee good video and audio
# response time.

# Display depth (bits)
DISPLAY_DEPTH       = 32 

# Check to see if the movie module is available.
try:
    import pygame.movie as movie
except ImportError:
    movie = None

# mpgPlayer Class
class mpgPlayer(pykPlayer):
    # Initialise the player instace
    def __init__(self, song, songDb, errorNotifyCallback=None, doneCallback=None):
        """The first parameter, song, may be either a pykdb.SongStruct
        instance, or it may be a filename. """

        pykPlayer.__init__(self, song, songDb, errorNotifyCallback, doneCallback)

        self.Movie = None

        manager.setCpuSpeed('mpg')

        manager.InitPlayer(self)
        manager.OpenDisplay(depth = DISPLAY_DEPTH)

        # Close the mixer while using Movie
        manager.CloseAudio()

        # Open the Movie module
        filepath = self.SongDatas[0].GetFilepath()
        if type(filepath) == str:
            filepath = filepath.encode(sys.getfilesystemencoding())
        self.Movie = pygame.movie.Movie(filepath)
        self.Movie.set_display(manager.display, (0, 0, manager.displaySize[0], manager.displaySize[1]))


    def doPlay(self):
        self.Movie.play()

    def doPause(self):
        self.Movie.pause()

    def doUnpause(self):
        self.Movie.play()

    def doRewind(self):
        self.Movie.stop()
        self.Movie.rewind()
            
    # Get the movie length (in seconds).
    def GetLength(self):
        return self.Movie.get_length()
        
    # Get the current time (in milliseconds).
    def GetPos(self):
        return (self.Movie.get_time() * 1000)

    def SetupOptions(self):
        """ Initialise and return optparse OptionParser object,
        suitable for parsing the command line options to this
        application. """

        parser = pykPlayer.SetupOptions(self, usage = "%prog [options] <mpg filename>")

        # Remove irrelevant options.
        parser.remove_option('--font-scale')
        
        return parser

    def shutdown(self):
        # This will be called by the pykManager to shut down the thing
        # immediately.
        if self.Movie:
            self.Movie.stop()
        # Must remove the object before using pygame.mixer module again
        self.Movie = None
        pykPlayer.shutdown(self)

    def handleEvent(self, event):
        if event.type == pygame.KEYDOWN and event.key == pygame.K_RETURN and (event.mod & (pygame.KMOD_LSHIFT | pygame.KMOD_RSHIFT | pygame.KMOD_LMETA | pygame.KMOD_RMETA)):
            # Shift/meta return: start/stop song.  Useful for keybinding apps.
            self.Close()
            return
        
        pykPlayer.handleEvent(self, event)

    # Internal. Only called by the pykManager.
    def doResize(self, newSize):
        # Resize the screen.
        self.Movie.set_display(manager.display, (0, 0, manager.displaySize[0], manager.displaySize[1]))

    # Internal. Only called by the pykManager.
    def doResizeBegin(self):
        # The Movie player must be paused while resizing otherwise we
        # get Xlib errors. pykmanager will call here before the resize
        # so that we can do it.
        if self.State == STATE_PLAYING:
            self.Movie.pause()

    # Internal. Only called by the pykManager.
    def doResizeEnd(self):
        # Called by pykmanager when resizing has finished.
        # We only play if it was playing in the first place.
        if self.State == STATE_PLAYING:
            self.Movie.play()

class externalPlayer(pykPlayer):

    """ This class is used to invoke an external command and wait for
    it to finish.  It is usually used to play a video file using an
    external player. """
    
    def __init__(self, song, songDb, errorNotifyCallback=None, doneCallback=None):
        """The first parameter, song, may be either a pykdb.SongStruct
        instance, or it may be a filename. """

        pykPlayer.__init__(self, song, songDb, errorNotifyCallback, doneCallback)

        self.Movie = None

        manager.setCpuSpeed('mpg')
        manager.InitPlayer(self)

        # Close the audio and the display
        manager.CloseAudio()
        manager.CloseDisplay()
        manager.CloseCPUControl()

        self.procReturnCode = None
        self.proc = None


    def doPlay(self):
        if self.procReturnCode != None:
            # The movie is done.
            self.__stop()

        if not self.proc:
            self.__start()

    def GetLength(self):
        # We cannot fetch the length from arbitrary external players.
        # Return zero-length.
        return 0

    def GetPos(self):
        # Use the default GetPos() which simply checks the time
        # since we started playing. This does not take account
        # for any fast-forward/rewind that may occur in the
        # external player, but we cannot support getting the
        # song position from arbitrary user-supplied players.
        return pykPlayer.GetPos(self)

    def doStuff(self):
        if self.procReturnCode != None:
            # The movie is done.
            self.__stop()
            self.Close()

        pykPlayer.doStuff(self)

    def __start(self):
        filepath = self.SongDatas[0].GetFilepath()

        external = manager.settings.MpgExternal
        if '%' in external:
            # Assume the filename parameter is embedded in the string.
            cmd = external % {
                'file' : filepath,
                }

        elif external:
            # No parameter appears to be present; assume the program
            # accepts the filename as the only parameter.
            cmd = [external, filepath]
            
        shell = True
        if env == ENV_WINDOWS:
            # Don't try to open the process under a "shell" in
            # Windows; that just breaks commands with spaces.
            shell = False

        assert self.procReturnCode == None
        sys.stdout.flush()
        self.proc = subprocess.Popen(cmd, shell = shell)
        if manager.settings.MpgExternalThreaded:
            # Wait for it to complete in a thread.
            self.thread = threading.Thread(target = self.__runThread)
            self.thread.start()
        else:
            # Don't wait for it in a thread; wait for it here.
            self.thread = None
            self.__runThread()

    def __stop(self):
        if self.thread:
            self.thread.join()
        self.proc = None
        self.procReturnCode = None
        self.thread = None
        

    def __runThread(self):
        """ This method runs in a sub-thread.  Its job is just to wait
        for the process to finish. """

        try:
            self.procReturnCode = self.proc.wait()
        except OSError:
            self.procReturnCode = -1
        
        

# Can be called from the command line with the MPG filepath as parameter
def main():
    player = mpgPlayer(None, None)
    player.Play()
    manager.WaitForPlayer()

if __name__ == "__main__":
    sys.exit(main())
    #import profile
    #result = profile.run('main()', 'pympg.prof')
    #sys.exit(result)
    



from . import pycdg
# from . import pycdgAux
# from . import pykar
# from . import pykconstants
# from . import pykdb
# from . import pykenv
# from . import pykmanager
# from . import pykplayer
# from . import pykversion
# from . import pympg



#
# Copyright (C) 2010  Kelvin Lawson (kelvinl@users.sourceforge.net)
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

""" This module defines various constants that are used throughout
this package. """

# Environment
ENV_WINDOWS = 1
ENV_POSIX = 2
ENV_OSX = 3
ENV_GP2X = 4
ENV_UNKNOWN = 5

# States
STATE_INIT          = 0
STATE_PLAYING       = 1
STATE_PAUSED        = 2
STATE_NOT_PLAYING   = 3
STATE_CLOSING       = 4
STATE_CLOSED        = 5
STATE_CAPTURING     = 6

# GP2X joystick button mappings
GP2X_BUTTON_UP            = (0)
GP2X_BUTTON_DOWN          = (4)
GP2X_BUTTON_LEFT          = (2)
GP2X_BUTTON_RIGHT         = (6)
GP2X_BUTTON_UPLEFT        = (1)
GP2X_BUTTON_UPRIGHT       = (7)
GP2X_BUTTON_DOWNLEFT      = (3)
GP2X_BUTTON_DOWNRIGHT     = (5)
GP2X_BUTTON_CLICK         = (18)
GP2X_BUTTON_A             = (12)
GP2X_BUTTON_B             = (13)
GP2X_BUTTON_X             = (14)
GP2X_BUTTON_Y             = (15)
GP2X_BUTTON_L             = (10)
GP2X_BUTTON_R             = (11)
GP2X_BUTTON_START         = (8) 
GP2X_BUTTON_SELECT        = (9)
GP2X_BUTTON_VOLUP         = (16)
GP2X_BUTTON_VOLDOWN       = (17)

# Left and top margins
Y_BORDER = 20
X_BORDER = 20



#!/usr/bin/env python

# pycdg - CDG/MP3+G Karaoke Player

# Copyright (C) 2010 Kelvin Lawson (kelvinl@users.sourceforge.net)
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# OVERVIEW
#
# pycdg is a CDG karaoke player which supports MP3+G and OGG+G tracks.
#
# The player uses the pygame library (www.pygame.org), and can therefore
# run on any operating system that runs pygame (currently Linux, Windows
# and OSX).
#
# You can use this file as a standalone player, or together with
# PyKaraoke. PyKaraoke provides a graphical user interface, playlists,
# searchable song database etc.
#
# For those writing a media player or similar project who would like
# CDG support, this module has been designed to be easily incorporated
# into such projects and is released under the LGPL.


# REQUIREMENTS
#
# pycdg requires the following to be installed on your system:
# . Python (www.python.org)
# . Pygame (www.pygame.org)

# . Numeric module (numpy.sourceforge.net) (actually, this is required
#    only if you do not use the compiled C low-level CDG implementation in
#    _pycdgAux.c)


# USAGE INSTRUCTIONS
#
# To start the player, pass the CDG filename/path on the command line:
#       python pycdg.py /songs/theboxer.cdg
#
# You can also incorporate a CDG player in your own projects by
# importing this module. The class cdgPlayer is exported by the
# module. You can import and start it as follows:
#   import pycdg
#   player = pycdg.cdgPlayer("/songs/theboxer.cdg")
#   player.Play()
# If you do this, you must also arrange to call pycdg.manager.Poll()
# from time to time, at least every 100 milliseconds or so, to allow
# the player to do its work.
#
# The class also exports Close(), Pause(), Rewind(), GetPos().
#
# There are two optional parameters to the initialiser, errorNotifyCallback
# and doneCallback:
#
# errorNotifyCallback, if provided, will be used to print out any error
# messages (e.g. song file not found). This allows the module to fit
# together well with GUI playlist managers by utilising the same GUI's
# error popup window mechanism (or similar). If no callback is provided,
# errors are printed to stdout. errorNotifyCallback should take one
# parameter, the error string, e.g.:
#   def errorPopup (ErrorString):
#       msgBox (ErrorString)
#
# doneCallback can be used to register a callback so that the player
# calls you back when the song is finished playing. The callback should
# take no parameters, e.g.:
#   def songFinishedCallback():
#       msgBox ("Song is finished")
#
# To register callbacks, pass the functions in to the initialiser:
#   cdgPlayer ("/songs/theboxer.cdg", errorPopup, songFinishedCallback)
# These parameters are optional and default to None.
#
# If the initialiser fails (e.g. the song file is not present), __init__
# raises an exception.


# IMPLEMENTATION DETAILS
#

# pycdg is implemented as a handful of python modules.  All of the CDG
# decoding is handled in the C module _pycdgAux.c, or in the
# equivalent (but slightly slower) pycdgAux.py if the C module is not
# available for some reason.  This Python implementation of
# pycdgAux.py uses the python Numeric module, which provides fast
# handling of the arrays of pixel data for the display.
#
# Audio playback and video display capabilities come from the pygame
# library.
#
# All of the information on the CDG file format was learned
# from the fabulous "CDG Revealed" tutorial at www.jbum.com.
#

# Previous implementations ran the player within a thread; this is no
# longer the case.  Instead, it is the caller's responsibility to call
# pycdg.manager.Poll() every once in a while to ensure that the player
# gets enough CPU time to do its work.  Ideally, this should be at
# least every 100 milliseconds or so to guarantee good video and audio
# response time.
#
# At each call to Poll(), the player checks the current time in the
# song. It reads the CDG file at the correct location for the current
# position of the song, and decodes the CDG commands stored there. If
# the CDG command requires a screen update, a local array of pixels is
# updated to reflect the new graphic information. Rather than update
# directly to the screen for every command, updates are cached and
# output to the screen a certain number of times per second
# (configurable). Performing the scaling and blitting required for
# screen updates might consume a lot of CPU horsepower, so we reduce
# the load further by dividing the screen into 24 segments. Only those
# segments that have changed are scaled and blitted. If the user
# resizes the window or we get a full-screen modification, the entire
# screen is updated, but during normal CD+G operation only a small
# number of segments are likely to be changed at update time.
#
# Here follows a description of the important data stored by
# the class:
#
# CdgPacketReader.__cdgColourTable[]
# Store the colours for each colour index (0-15).
# These are set using the load colour look up table commands.
#
# CdgPacketReader.__cdgSurfarray[300][216]
# Surfarray object containing pixel colours for the full 300x216 screen.
# The border area is not actually displayed on the screen, however we
# need to store the pixel colours there as they are set when Scroll
# commands are used. This stores the actual pygame colour value, not
# indeces into our colour table.
#
# CdgPacketReader.__cdgPixelColours[300][216]
# Store the colour index for every single pixel. The values stored
# are indeces into our colour table, rather than actual pygame
# colour representations. It's unfortunate that we need to store
# all this data, when in fact the pixel colour is available from
# cdgSurfarray, but we need it for the Tile Block XOR command.
# The XOR command performs an XOR of the colour index currently
# at the pixel, with the new colour index. We therefore need to
# know the actual colour index at that pixel - we can't do a
# get_at() on the screen, or look in cdgSurfarray, and map the RGB
# colour back to a colour index because some CDG files have the
# same colour in two places in the table, making it impossible to
# determine which index is relevant for the XOR.
#
# CdgPacketReader.__cdgPresetColourIndex
# Preset Colour (index into colour table)
#
# CdgPacketReader.__cdgPresetColourIndex
# Border Colour (index into colour table)
#
# CdgPacketReader.__updatedTiles
# Bitmask to mark which screen segments have been updated.
# This is used to reduce the amount of effort required in
# scaling the output video. This is an expensive operation
# which must be done for every screen update so we divide
# the screen into 24 segments and only update those segments
# which have actually been updated.

from .pykconstants import *
from .pykplayer import pykPlayer
from .pykenv import env
from .pykmanager import manager
import sys, pygame, os, string, math

# Import the optimised C version if available, or fall back to Python
try:
    import _pycdgAux as aux_c
except ImportError:
    aux_c = None

try:
    from . import pycdgAux as aux_python
except ImportError as ex:
    print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!', ex)
    aux_python = None

CDG_DISPLAY_WIDTH   = 288
CDG_DISPLAY_HEIGHT  = 192

# Screen tile positions
# The viewable area of the screen (294x204) is divided into 24 tiles
# (6x4 of 49x51 each). This is used to only update those tiles which
# have changed on every screen update, thus reducing the CPU load of
# screen updates. A bitmask of tiles requiring update is held in
# cdgPlayer.UpdatedTiles.  This stores each of the 4 columns in
# separate bytes, with 6 bits used to represent the 6 rows.
TILES_PER_ROW           = 6
TILES_PER_COL           = 4
TILE_WIDTH              = CDG_DISPLAY_WIDTH / TILES_PER_ROW
TILE_HEIGHT             = CDG_DISPLAY_HEIGHT / TILES_PER_COL

# cdgPlayer Class
class cdgPlayer(pykPlayer):
    # Initialise the player instace
    def __init__(self, song, songDb, errorNotifyCallback=None,
                 doneCallback=None):
        """The first parameter, song, may be either a pykdb.SongStruct
        instance, or it may be a filename. """

        pykPlayer.__init__(self, song, songDb, errorNotifyCallback, doneCallback)

        # With the nomusic option no music will be played.
        soundFileData = None
        if not manager.options.nomusic:
            # Check for a matching mp3 or ogg file.  Check extensions
            # in the following order.
            validexts = [
                '.wav', '.ogg', '.mp3'
            ]

            for ext in validexts:
                for data in self.SongDatas:
                    if data.Ext == ext:
                        # Found a match!
                        soundFileData = data
                        break
                if soundFileData:
                    break

            if not soundFileData:
                ErrorString = "There is no mp3 or ogg file to match " + self.Song.DisplayFilename
                self.ErrorNotifyCallback (ErrorString)
                raise 'NoSoundFile'

        self.cdgFileData = self.SongDatas[0]
        self.soundFileData = soundFileData
        self.soundLength = 0

        # Handle a bug in pygame (pre-1.7) which means that the position
        # timer carries on even when the song has been paused.
        self.pauseOffsetTime = 0

        manager.InitPlayer(self)
        manager.OpenDisplay()
        manager.surface.fill((0, 0, 0))

        # A working surface for blitting tiles, one at a time.
        self.workingTile = pygame.Surface((TILE_WIDTH, TILE_HEIGHT),
                                          0, manager.surface)

        # A surface that contains the set of all tiles as they are to
        # be assembled onscreen.  This surface is kept at the original
        # scale, then zoomed to display size.  It is only used if
        # settings.CdgZoom == 'soft'.
        self.workingSurface = pygame.Surface((CDG_DISPLAY_WIDTH, CDG_DISPLAY_HEIGHT),
                                             pygame.HWSURFACE,
                                             manager.surface)

        self.borderColour = None
        self.computeDisplaySize()

        aux = aux_c
        if not aux or not manager.settings.CdgUseC:
            print("Using Python implementation of CDG interpreter.")
            aux = aux_python

        # Open the cdg and sound files
        self.packetReader = aux.CdgPacketReader(self.cdgFileData.GetData(), self.workingTile)
        manager.setCpuSpeed('cdg')

        if self.soundFileData:
            # Play the music normally.
            audioProperties = None
            if manager.settings.UseMp3Settings:
                audioProperties = self.getAudioProperties(self.soundFileData)
            if audioProperties == None:
                audioProperties = (None, None, None)
            try:
                manager.OpenAudio(*audioProperties)
                audio_path = self.soundFileData.GetFilepath()
                if type(audio_path) == str:
                    audio_path = audio_path.encode(sys.getfilesystemencoding())
                pygame.mixer.music.load(audio_path)
            except:
                self.Close()
                raise

            # Set an event for when the music finishes playing
            pygame.mixer.music.set_endevent(pygame.USEREVENT)

            # Account for the size of the playback buffer in the lyrics
            # display.  Assume that the buffer will be mostly full.  On a
            # slower computer that's struggling to keep up, this may not
            # be the right amount of delay, but it should usually be
            # pretty close.
            self.InternalOffsetTime = -manager.GetAudioBufferMS()

        else:
            # Don't play anything.
            self.InternalOffsetTime = 0

        # Set the CDG file at the beginning
        self.cdgReadPackets = 0
        self.cdgPacketsDue = 0
        self.LastPos = self.curr_pos = 0

        # Some session-wide constants.
        self.ms_per_update = (1000.0 / manager.options.fps)

    def doPlay(self):
        if self.soundFileData:
            pygame.mixer.music.play()

    # Pause the song - Use Pause() again to unpause
    def doPause(self):
        if self.soundFileData:
            pygame.mixer.music.pause()
            self.PauseStartTime = self.GetPos()

    def doUnpause(self):
        if self.soundFileData:
            self.pauseOffsetTime = self.pauseOffsetTime + (self.GetPos() - self.PauseStartTime)
            pygame.mixer.music.unpause()

    # you must call Play() to restart. Blocks until pygame is initialised
    def doRewind(self):
        # Reset the state of the packet-reading thread
        self.cdgReadPackets = 0
        self.cdgPacketsDue = 0
        self.LastPos = 0
        # No need for the Pause() fix anymore
        self.pauseOffsetTime = 0
        # Move file pointer to the beginning of the file
        self.packetReader.Rewind()

        if self.soundFileData:
            # Actually stop the audio
            pygame.mixer.music.rewind()
            pygame.mixer.music.stop()

    def GetLength(self):
        """Give the number of seconds in the song."""
        return self.soundLength

    # Get the current time (in milliseconds). Blocks if pygame is
    # not initialised yet.
    def GetPos(self):
        if self.soundFileData:
            return pygame.mixer.music.get_pos()
        else:
            return pykPlayer.GetPos(self)

    def SetupOptions(self):
        """ Initialise and return optparse OptionParser object,
        suitable for parsing the command line options to this
        application. """

        parser = pykPlayer.SetupOptions(self, usage = "%prog [options] <CDG file>")

        # Remove irrelevant options.
        parser.remove_option('--font-scale')

        return parser


    def shutdown(self):
        # This will be called by the pykManager to shut down the thing
        # immediately.
        if self.soundFileData:
            if manager.audioProps:
                pygame.mixer.music.stop()

        # Make sure our surfaces are deallocated before we call up to
        # CloseDisplay(), otherwise bad things can happen.
        self.workingSurface = None
        self.workingTile = None
        self.packetReader = None
        pykPlayer.shutdown(self)

    def doStuff(self):
        pykPlayer.doStuff(self)

        # Check whether the songfile has moved on, if so
        # get the relevant CDG data and update the screen.
        if self.State == STATE_PLAYING or self.State == STATE_CAPTURING:
            self.curr_pos = self.GetPos() + self.InternalOffsetTime + manager.settings.SyncDelayMs - self.pauseOffsetTime

            self.cdgPacketsDue = int((self.curr_pos * 300) / 1000)
            numPackets = self.cdgPacketsDue - self.cdgReadPackets
            if numPackets > 0:
                if not self.packetReader.DoPackets(numPackets):
                    # End of file.
                    #print "End of file on cdg."
                    self.Close()
                self.cdgReadPackets += numPackets

            # Check if any screen updates are now due.
            if (self.curr_pos - self.LastPos) > self.ms_per_update:
                self.cdgDisplayUpdate()
                self.LastPos = self.curr_pos

    def handleEvent(self, event):
        if event.type == pygame.KEYDOWN and event.key == pygame.K_RETURN and (event.mod & (pygame.KMOD_LSHIFT | pygame.KMOD_RSHIFT | pygame.KMOD_LMETA | pygame.KMOD_RMETA)):
            # Shift/meta return: start/stop song.  Useful for keybinding apps.
            self.Close()
            return
        
        pykPlayer.handleEvent(self, event)

    def doResize(self, newSize):
        self.computeDisplaySize()

        if self.borderColour != None:
            manager.surface.fill(self.borderColour)

        self.packetReader.MarkTilesDirty()

    def computeDisplaySize(self):
        """ Figures out what scale and placement to use for blitting
        tiles to the screen.  This must be called at startup, and
        whenever the window size changes. """

        winWidth, winHeight = manager.displaySize

        # Compute an appropriate uniform scale to letterbox the image
        # within the window
        scale = min(float(winWidth) / CDG_DISPLAY_WIDTH,
                    float(winHeight) / CDG_DISPLAY_HEIGHT)
        if manager.settings.CdgZoom == 'none':
            scale = 1
        elif manager.settings.CdgZoom == 'int':
            if scale < 1:
                scale = 1.0/math.ceil(1.0/scale)
            else:
                scale = int(scale)
        self.displayScale = scale

        scaledWidth = int(scale * CDG_DISPLAY_WIDTH)
        scaledHeight = int(scale * CDG_DISPLAY_HEIGHT)

        if manager.settings.CdgZoom == 'full':
            # If we are allowing non-proportional scaling, allow
            # scaledWidth and scaledHeight to be independent.
            scaledWidth = winWidth
            scaledHeight = winHeight

        # And the center of the display after letterboxing.
        self.displayRowOffset = (winWidth - scaledWidth) / 2
        self.displayColOffset = (winHeight - scaledHeight) / 2

        # Calculate the scaled width and height for each tile
        if manager.settings.CdgZoom == 'soft':
            self.displayTileWidth = CDG_DISPLAY_WIDTH / TILES_PER_ROW
            self.displayTileHeight = CDG_DISPLAY_HEIGHT / TILES_PER_COL
        else:
            self.displayTileWidth = scaledWidth / TILES_PER_ROW
            self.displayTileHeight = scaledHeight / TILES_PER_COL

    def getAudioProperties(self, soundFileData):
        """ Attempts to determine the samplerate, etc., from the
        specified filename.  It would be nice to know this so we can
        configure the audio hardware to the same properties, to
        minimize run-time resampling. """

        # Ideally, SDL would tell us this (since it knows!), but
        # SDL_mixer doesn't provide an interface to query this
        # information, so we have to open the soundfile separately and
        # try to figure it out ourselves.

        audioProperties = None
        if soundFileData.Ext == '.mp3':
            audioProperties = self.getMp3AudioProperties(soundFileData)

        return audioProperties

    def getMp3AudioProperties(self, soundFileData):
        """ Attempts to determine the samplerate, etc., from the
        specified filename, which is known to be an mp3 file. """

        # Hopefully we have Mutagen available to pull out the song length
        try:
            import mutagen.mp3
        except:
            print("Mutagen not available, will not be able to determine extra MP3 information.")
            self.soundLength = 0
            return None

        # Open the file with mutagen
        m = mutagen.mp3.MP3(soundFileData.GetFilepath())

        # Pull out the song length
        self.soundLength = m.info.length

        # Get the number of channels, mode field of 00 or 01 indicate stereo
        if m.info.mode < 2:
            channels = 2
        else:
            channels = 1

        # Put the channels and sample rate together in a tuple and return
        audioProperties = (m.info.sample_rate, -16, channels)
        return audioProperties

    # Actually update/refresh the video output
    def cdgDisplayUpdate(self):
        # This routine is responsible for taking the unscaled output
        # pixel data from self.cdgSurfarray, scaling it and blitting
        # it to the actual display surface. The viewable area of the
        # unscaled surface is 294x204 pixels.  Because scaling and
        # blitting are heavy operations, we divide the screen into 24
        # tiles, and only scale and blit those tiles which have been
        # updated recently.  The CdgPacketReader class
        # (self.packetReader) is responsible for keeping track of
        # which areas of the screen have been modified.

        # There are four different approaches for blitting tiles onto
        # the display:

        # settings.CdgZoom == 'none':
        #   No scaling.  The CDG graphics are centered within the
        #   display.  When a tile is dirty, it is blitted directly to
        #   manager.surface.  After all dirty tiles have been blitted,
        #   we then use display.update to flip only those rectangles
        #   on the screen that have been blitted.

        # settings.CdgZoom = 'quick':
        #   Trivial scaling.  Similar to 'none', but each tile is
        #   first scaled to its target scale using
        #   pygame.transform.scale(), which is quick but gives a
        #   pixelly result.  The scaled tile is then blitted to
        #   manager.surface.

        # settings.CdgZoom = 'int':
        #   The same as 'quick', but the scaling is constrained to be
        #   an integer multiple or divisor of its original size, which
        #   may reduce artifacts somewhat.

        # settings.CdgZoom = 'full':
        #   The same as 'quick', but the scaling is allowed to
        #   completely fill the window in both x and y, regardless of
        #   aspect ratio constraints.

        # settings.CdgZoom = 'soft':
        #   Antialiased scaling.  We blit all tiles onto
        #   self.workingSurface, which is maintained as the non-scaled
        #   version of the CDG graphics, similar to 'none'.  Then,
        #   after all dirty tiles have been blitted to
        #   self.workingSurface, we use pygame.transform.rotozoom() to
        #   make a nice, antialiased scaling of workingSurface to
        #   manager.surface, and then flip the whole display.  (We
        #   can't scale and blit the tiles one a time in this mode,
        #   since that introduces artifacts between the tile edges.)
        borderColour = self.packetReader.GetBorderColour()
        if borderColour != self.borderColour:
            # When the border colour changes, blit the whole screen
            # and redraw it.
            self.borderColour = borderColour
            if borderColour != None:
                manager.surface.fill(borderColour)
                self.packetReader.MarkTilesDirty()

        dirtyTiles = self.packetReader.GetDirtyTiles()
        if not dirtyTiles:
            # If no tiles are dirty, don't bother.
            return

        # List of update rectangles (in scaled output window)
        rect_list = []

        # Scale and blit only those tiles which have been updated
        for row, col in dirtyTiles:
            self.packetReader.FillTile(self.workingTile, row, col)

            if manager.settings.CdgZoom == 'none':
                # The no-scale approach.
                rect = pygame.Rect(self.displayTileWidth * row + self.displayRowOffset,
                                   self.displayTileHeight * col + self.displayColOffset,
                                   self.displayTileWidth, self.displayTileHeight)
                manager.surface.blit(self.workingTile, rect)
                rect_list.append(rect)

            elif manager.settings.CdgZoom == 'soft':
                # The soft-scale approach.
                self.workingSurface.blit(self.workingTile, (self.displayTileWidth * row, self.displayTileHeight * col))

            else:
                # The quick-scale approach.
                self.displayTileWidth = int(self.displayTileWidth)
                self.displayTileHeight = int(self.displayTileHeight)
                scaled = pygame.transform.scale(self.workingTile, (self.displayTileWidth,self.displayTileHeight))
                rect = pygame.Rect(self.displayTileWidth * row + self.displayRowOffset,
                                   self.displayTileHeight * col + self.displayColOffset,
                                   self.displayTileWidth, self.displayTileHeight)
                manager.surface.blit(scaled, rect)
                rect_list.append(rect)

        if manager.settings.CdgZoom == 'soft':
            # Now scale and blit the whole screen.
            scaled = pygame.transform.rotozoom(self.workingSurface, 0, self.displayScale)
            manager.surface.blit(scaled, (self.displayRowOffset, self.displayColOffset))
            manager.Flip()
        elif len(rect_list) < 24:
            # Only update those areas which have changed
            if manager.display:
                pygame.display.update(rect_list)
        else:
            manager.Flip()

def defaultErrorPrint(ErrorString):
    print (ErrorString)

# Can be called from the command line with the CDG filepath as parameter
def main():
    player = cdgPlayer(None, None)
    player.Play()
    manager.WaitForPlayer()

if __name__ == "__main__":
    sys.exit(main())
    #import profile
    #result = profile.run('main()', 'pycdg.prof')
    #sys.exit(result)




#******************************************************************************
#****                                                                      ****
#**** Copyright (C) 2010  Kelvin Lawson (kelvinl@users.sourceforge.net)    ****
#**** Copyright (C) 2010  PyKaraoke Development Team                       ****
#****                                                                      ****
#**** This library is free software; you can redistribute it and/or        ****
#**** modify it under the terms of the GNU Lesser General Public           ****
#**** License as published by the Free Software Foundation; either         ****
#**** version 2.1 of the License, or (at your option) any later version.   ****
#****                                                                      ****
#**** This library is distributed in the hope that it will be useful,      ****
#**** but WITHOUT ANY WARRANTY; without even the implied warranty of       ****
#**** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    ****
#**** Lesser General Public License for more details.                      ****
#****                                                                      ****
#**** You should have received a copy of the GNU Lesser General Public     ****
#**** License along with this library; if not, write to the                ****
#**** Free Software Foundation, Inc.                                       ****
#**** 59 Temple Place, Suite 330                                           ****
#**** Boston, MA  02111-1307  USA                                          ****
#******************************************************************************

""" This module provides support for the PyKaraoke song database, as
well as the user's settings file. """

import pygame
from .pykconstants import *
from .pykenv import env
from . import pykar
from . import pycdg
from . import pympg
import os, pickle, zipfile, codecs, sys, time
import types
from io import StringIO
try:
    from hashlib import md5
except ImportError:
    from md5 import md5

# The amount of time to wait, in milliseconds, before yielding to the
# app for windowing updates during a long update process.
YIELD_INTERVAL = 1000

# The maximum number of zip files we will attempt to store in our zip
# file cache.
MAX_ZIP_FILES = 10

# Increment this version number whenever the settings version changes
# (which may not necessarily change with each PyKaraoke release).
# This will force users to re-enter their configuration information.
SETTINGS_VERSION = 6

# Increment this version number whenever the database version changes
# (which will also hopefully be infrequently).
DATABASE_VERSION = 2

class AppYielder:
    """ This is a simple class that knows how to yield control to the
    windowing system every once in a while.  It is passed to functions
    like SearchDatabase and BuildSearchDatabase--tasks which might
    take a while to perform.

    This class is just an abstract base class and does nothing.  Apps
    should subclass it and override Yield() to make any use from
    it. """

    def __init__(self):
        self.lastYield = pygame.time.get_ticks()

    def ConsiderYield(self):
        now = pygame.time.get_ticks()
        if now - self.lastYield >= YIELD_INTERVAL:
            self.Yield()
            self.lastYield = now

    def Yield(self):
        """ Override this method to actually yield control to the
        windowing system. """
        pass

class BusyCancelDialog:

    """This class implements a busy dialog to show a task is
    progressing, and it includes a cancel button the user might click
    on to interrupt the task.  This is just an abstract base class and
    does nothing.  Apps should subclass from it. """

    def __init__(self):
        self.Clicked = False

    def Show(self):
        pass

    def SetProgress(self, label, progress):
        pass

    def Destroy(self):
        pass

class SongData:
    """This class is returned by SongStruct.GetSongDatas(), below.  It
    represents either a song file that exists on disk (and must still
    be read), or it is a song file that was found in a zip archive
    (and its data is available now)."""

    def __init__(self, filename, data):
        self.filename = filename
        self.tempFilename = None
        self.data = data
        self.Ext = os.path.splitext(filename)[1].lower()

        # By convention, if data is passed as None to the constructor,
        # that means this file is a true file that exists on disk.  On
        # the other hand, if data is not None, then this is not a true
        # file, and data contains its contents.
        self.trueFile = (data == None)

    def GetData(self):
        """Returns the actual data of the file.  If the file has not
        yet been read, this will read it and return the data."""

        if self.data != None:
            # The file has already been read; return that data.
            return self.data

        # The file has not yet been read
        self.data = open(self.filename, 'rb').read()
        return self.data

    def GetFilepath(self):
        """Returns a full pathname to the file.  If the file does not
        exist on disk, this will write it to a temporary file and
        return the name of that file."""

        if self.trueFile:
            # The file exists on disk already; just return its
            # pathname.
            return self.filename

        if not self.tempFilename:
            # The file does not exist on disk already; we have to write it
            # to a temporary file.
            prefix = globalSongDB.CreateTempFileNamePrefix()
            basename = os.path.basename(self.filename)
            # Add the tempfile prefix as well as the time to make the
            # filename unique. This works around a bug
            # in pygame.mixer.music on Windows which stops us deleting
            # temp files until another song is loaded, preventing the
            # same song from being played twice in a row.
            self.tempFilename = prefix + str(time.time()) + basename
            open(self.tempFilename, 'wb').write(self.data)

        return self.tempFilename


# This functor is declared globally.  It is assigned by
# SongDB.SelectSort(), so we can use bisect to search through
# the list also.  This is an ugly hack around the fact that bisect has
# no facility to receive a key parameter, like sort does.
fileSortKey = None

class SongStruct:
    """ This corresponds to a single song file entry, e.g. a .kar
    file, or a .mp3/.cdg filename pair.  The file might correspond to
    a physical file on disk, or to a file within a zip file. """

    # Type codes.
    T_KAR = 0
    T_CDG = 1
    T_MPG = 2

    def __init__(self, Filepath, settings,
                 Title = None, Artist = None, ZipStoredName = None, DatabaseAdd = False):
        self.Filepath = Filepath    # Full path to file or ZIP file
        self.ZipStoredName = ZipStoredName # Filename stored in ZIP

        # Assume there will be no title/artist info found
        self.Title = Title or ''    # (optional) Title for display in playlist
        self.Artist = Artist or ''  # (optional) Artist for display
        self.Disc = '' # (optional) Disc for display
        self.Track = -1 # (optional) Track for display

        # Check to see if we are deriving song information from the filename
        if settings.CdgDeriveSongInformation:
            try:
                self.Title = self.ParseTitle(Filepath, settings)    # Title for display in playlist
                self.Artist = self.ParseArtist(Filepath, settings)  # Artist for display
                self.Disc = self.ParseDisc(Filepath, settings)      # Disc for display
                self.Track = self.ParseTrack(Filepath, settings)     # Track for display
            except:
                # Filename did not match requested scheme, set the title to the filepath
                # so that the structure is still created, but without any additional info
                #print "Filename format does not match requested scheme: %s" % Filepath
                self.Title = os.path.basename(Filepath)

                # If this SongStruct is being used to add to the database, and we are
                # configured to exclude non-matching files, raise an exception. Otherwise
                # allow it through. For database adds where we are not excluding such
                # files the song will still be added to the database. For non-database
                # adds we don't care anyway, we just want a SongStruct for passing around.
                if DatabaseAdd and settings.ExcludeNonMatchingFilenames:
                    raise KeyError("Excluding non-matching file: %s" % self.Title)

        # This is a list of other song files that share the same
        # artist and title data.
        self.sameSongs = []

        # This is a pointer to the TitleStruct object that defined
        # this song file, or None if it was not defined.
        self.titles = None

        # If the file ends in '.', assume we got it via tab-completion
        # on a filename, and it really is meant to end in '.cdg'.
        if self.Filepath != '' and self.Filepath[-1] == '.':
            self.Filepath += 'cdg'

        if ZipStoredName:
            self.DisplayFilename = os.path.basename(ZipStoredName)
            if isinstance(self.DisplayFilename, bytes):
                self.DisplayFilename = self.DisplayFilename.decode(settings.ZipfileCoding)
        else:
            self.DisplayFilename = os.path.basename(Filepath)
            if isinstance(self.DisplayFilename, bytes):
                self.DisplayFilename = self.DisplayFilename.decode(settings.FilesystemCoding)

        # Check the file type based on extension.
        self.Type = None
        ext = os.path.splitext(self.DisplayFilename)[1].lower()
        if ext in settings.KarExtensions:
            self.Type = self.T_KAR
        elif ext in settings.CdgExtensions:
            self.Type = self.T_CDG
        elif ext in settings.MpgExtensions:
            self.Type = self.T_MPG
            if ext == '.mpg' or ext == '.mpeg':
                self.MpgType = 'mpg'
            else:
                self.MpgType = ext[1:]

    def ParseTitle(self, filepath, settings):
        """ Parses the file path and returns the title of the song. If the filepath cannot be parsed a KeyError exception is thrown. If the settings contains a file naming scheme that we do not support a KeyError exception is thrown."""
        title = ''
        # Make sure we are to parse information
        if settings.CdgDeriveSongInformation:
            if settings.CdgFileNameType == 0: # Disc-Track-Artist-Title.Ext
                # Make sure we can parse the filepath
                if len(filepath.split("-")) == 4:
                    title = filepath.split("-")[3] # Find the Title in the filename
                else:
                    raise KeyError("Invalid type for file: %s!" % filepath)
            elif settings.CdgFileNameType == 1: # DiscTrack-Artist-Title.Ext
                # Make sure we can parse the filepath
                if len(filepath.split("-")) == 3:
                    title = filepath.split("-")[2] # Find the Title in the filename
                else:
                    raise KeyError("Invalid type for file: %s!" % filepath)
            elif settings.CdgFileNameType == 2: # Disc-Artist-Title.Ext
                # Make sure we can parse the filepath
                if len(filepath.split("-")) == 3:
                    title = filepath.split("-")[2] # Find the Title in the filename
                else:
                    raise KeyError("Invalid type for file: %s!" % filepath)
            elif settings.CdgFileNameType == 3: # Artist-Title.Ext
                # Make sure we can parse the filepath
                if len(filepath.split("-")) == 2:
                    title = filepath.split("-")[1] # Find the Title in the filename
                else:
                    raise KeyError("Invalid type for file: %s!" % filepath)
            else:
                raise KeyError("File name type is invalid!")
            # Remove the first and last space
            title = title.strip(" ")
            # Remove the filename extension
            title = os.path.splitext(title)[0]
        #print "Title parsed: %s" % title
        return title

    def ParseArtist(self, filepath, settings):
        """ Parses the filepath and returns the artist of the song. """
        artist = ''
        # Make sure we are to parse information
        if settings.CdgDeriveSongInformation:
            if settings.CdgFileNameType == 0: # Disc-Track-Artist-Title.Ext
                artist = filepath.split("-")[2] # Find the Artist in the filename
            elif settings.CdgFileNameType == 1: # DiscTrack-Artist-Title.Ext
                artist = filepath.split("-")[1] # Find the Artist in the filename
            elif settings.CdgFileNameType == 2: # Disc-Artist-Title.Ext
                artist = filepath.split("-")[1] # Find the Artist in the filename
            elif settings.CdgFileNameType == 3: # Artist-Title.Ext
                artist = filepath.split("-")[0] # Find the Artist in the filename
                artist = os.path.basename(artist)
            else:
                raise KeyError("File name type is invalid!")
            # Remove the first and last space
            artist = artist.strip(" ")
        #print "Artist parsed: %s" % artist
        return artist

    def ParseDisc(self, filepath, settings):
        """ Parses the filepath and returns the disc name of the song. """
        disc = ''
        # Make sure we are to parse information
        if settings.CdgDeriveSongInformation:
            if settings.CdgFileNameType == 0: # Disc-Track-Artist-Title.Ext
                disc = filepath.split("-")[0] # Find the Disc in the filename
            elif settings.CdgFileNameType == 1: # DiscTrack-Artist-Title.Ext
                disc = filepath.mid(0, filepath.length - 2) # Find the Disc in the filename
            elif settings.CdgFileNameType == 2: # Disc-Artist-Title.Ext
                disc = filepath.split("-")[0] # Find the Disc in the filename
            elif settings.CdgFileNameType == 3: # Artist-Title.Ext
                disc = ''
            else:
                raise KeyError("File name type is invalid!")
            # Remove the first and last space
            disc = disc.strip(" ")
            # Remove the filename path
            disc = os.path.basename(disc)
        #print "Disc parsed: %s" % disc
        return disc

    def ParseTrack(self, filepath, settings):
        """ Parses the file path and returns the track for the song. """
        track = ''
        # Make sure we are to parse information
        if settings.CdgDeriveSongInformation:
            if settings.CdgFileNameType == 0: # Disc-Track-Artist-Title.Ext
                track = filepath.split("-")[1] # Find the Track in the filename
            elif settings.CdgFileNameType == 1: # DiscTrack-Artist-Title.Ext
                track = filepath.mid(filepath.length - 2, 2) # Find the Track in the filename
            elif settings.CdgFileNameType == 2: # Disc-Artist-Title.Ext
                track = ''
            elif settings.CdgFileNameType == 3: # Artist-Title.Ext
                track = ''
            else:
                raise KeyError("File name type is invalid!")
            # Remove the first and last space
            #track = track.strip(" ")
        #print "Track parsed: %s" % track
        return track

    def MakeSortKey(self, str):
        """ Returns a suitable key to use for sorting, by lowercasing
        and removing articles from the indicated string. """
        str = str.strip().lower()
        if str:
            # Remove a leading parenthetical phrase.
            if str[0] == '(':
                rparen = str.index(')')
                if rparen != ')':
                    str = str[rparen + 1:].strip()

        if str:
            # Remove a leading article.
            firstWord = str.split()[0]
            if firstWord in ['a', 'an', 'the']:
                str = str[len(firstWord):].strip()
                
        return str

    def MakePlayer(self, songDb, errorNotifyCallback, doneCallback):
        """Creates and returns a player of the appropriate type to
        play this file, if possible; or returns None if the file
        cannot be played (in which case, the errorNotifyCallback will
        have already been called with the error message). """

        settings = songDb.Settings
        constructor = None

        if self.Type == self.T_CDG:
            constructor = pycdg.cdgPlayer
        elif self.Type == self.T_KAR:
            constructor = pykar.midPlayer
        elif self.Type == self.T_MPG:
            if self.MpgType == 'mpg' and settings.MpgNative and pympg.movie:
                # Mpg files can be played internally.
                constructor = pympg.mpgPlayer
            else:
                # Other kinds of movies require an external player.
                constructor = pympg.externalPlayer
        else:
            ext = os.path.splitext(self.DisplayFilename)[1]
            errorNotifyCallback("Unsupported file format " + ext)
            return None

        # Try to open the song file.
        try:
            player = constructor(self, songDb, errorNotifyCallback,
                                 doneCallback)
        except:
            errorNotifyCallback("Error opening file.\n%s\n%s" % (sys.exc_info()[0], sys.exc_info()[1]))
            return None

        return player

    def GetSongDatas(self):
        """Returns a list of SongData objects; see SongData.

        Usually there is only one element in the list: the file named
        by this SongStruct.  In the case of .cdg files, however, there
        may be more tuples; the first tuple will be the file named by
        this SongStruct, and the remaining tuples will correspond to
        other files with the same basenames but different extensions
        (so that the .mp3 or .ogg associated with a cdg file may be
        recovered)."""

        songDatas = []

        if not self.Filepath:
            return songDatas

        if not os.path.exists(self.Filepath):
            error = 'No such file: %s' % (self.Filepath)
            raise ValueError(error)

        dir = os.path.dirname(self.Filepath)
        if dir == "":
            dir = "."
        root, ext = os.path.splitext(self.Filepath)
        prefix = os.path.basename(root + ".")

        if self.ZipStoredName:
            # It's in a ZIP file; unpack it.
            zip = globalSongDB.GetZipFile(self.Filepath)
            filelist = [self.ZipStoredName]

            root, ext = os.path.splitext(self.ZipStoredName)
            prefix = os.path.basename(root + ".")

            if self.Type == self.T_CDG:
                # In addition to the .cdg file, we also have to get
                # out the mp3/ogg/whatever audio file that comes with
                # the .cdg file.  Just extract out any files that have
                # the same basename.
                for name in zip.namelist():
                    if name != self.ZipStoredName and name.startswith(prefix):
                        filelist.append(name)

            # We'll continue looking for matching files outside the
            # zip, too.

            for file in filelist:
                try:
                    data = zip.read(file)
                    songDatas.append(SongData(file, data))
                except:
                    print("Error in ZIP containing " + file)
        else:
            # A non-zipped file; this is an easy case.
            songDatas.append(SongData(self.Filepath, None))

        if self.Type == self.T_CDG:
            # In addition to the .cdg file, we also have to find the
            # mp3/ogg/whatever audio file that comes with the .cdg
            # file, just as above, when we were pulling them out of
            # the zip file.  This time we are just looking for loose
            # files on the disk.
            for file in os.listdir(dir):
                # Handle potential byte-strings with invalid characters
                # that startswith() will not handle.
                try:
                    file = str(file)
                except UnicodeDecodeError:
                    file = file.decode("ascii", "replace")
                try:
                    prefix = str(prefix)
                except UnicodeDecodeError:
                    prefix = prefix.decode("ascii", "replace")

                # Check for a file which matches the prefix
                if file.startswith(prefix):
                    path = os.path.join(dir, file)
                    if path != self.Filepath:
                        songDatas.append(SongData(path, None))

        # Now we've found all the matching files.
        return songDatas

    def getTextColour(self, selected):
        """ Returns a suitable colour to use when rendering the text
        of this song line in the pykaraoke_mini song index. """

        if selected:
            fg = (255, 255, 255)

        else:
            # Determine the color of the text.
            fg = (180, 180, 180)
            if self.Type == self.T_KAR:
                # Midi file: color it red.
                fg = (180, 72, 72)

            elif self.Type == self.T_CDG:
                # CDG+MP3: color it blue.
                fg = (72, 72, 180)

            elif self.Type == self.T_MPG:
                # MPEG file: color it yellow.
                fg = (180, 180, 72)

        return fg


    def getBackgroundColour(self, selected):
        """ Returns a suitable colour to use when rendering the
        background of this song line in the pykaraoke_mini song
        index. """

        if not selected:
            bg = (0, 0, 0)

        else:
            if self.Type == self.T_KAR:
                # Midi file: color it red.
                bg = (120, 0, 0)

            elif self.Type == self.T_CDG:
                # CDG+MP3: color it blue.
                bg = (0, 0, 120)

            elif self.Type == self.T_MPG:
                # MPEG file: color it yellow.
                bg = (120, 120, 0)

        return bg

    def getDisplayFilenames(self):
        """ Returns the list of all of the filenames that share the
        same artist/title with this song file.  The list is formatted
        as a single comma-delimited string. """

        if self.sameSongs:
            return ', '.join([f.DisplayFilename for f in self.sameSongs])
        return self.DisplayFilename

    def getTypeSort(self):
        """Defines a sorting order by type, for sorting the sameSongs
        list. """

        # We negate self.Type, so that the sort order is: mpg, cdg,
        # kar.  This means that MPG files have priority over CDG which
        # have priority over KAR, for the purposes of coloring the
        # files in the mini index.
        return (-self.Type, self.DisplayFilename)

    def getMarkKey(self):
        """ Returns a key for indexing into markedSongs, for uniquely
        identifying this particular song file. """
        return (self.Filepath, self.ZipStoredName)

    def __cmp__(self, other):
        """Define a sorting order between SongStruct objects.  This is
        used in bisect, to quickly search for a SongStruct in a sorted
        list.  It relies on fileSortKey (above) having being filled in
        already. """
        global fileSortKey

        a = fileSortKey(self)
        b = fileSortKey(other)
        if a == b:
            return cmp(id(self), id(other))
        if a < b:
            return -1
        return 1

class TitleStruct:
    """ This represents a single titles.txt file.  Its filename is
    preserved so it can be rewritten later, to modify a title and/or
    artist associated with a song. """

    def __init__(self,  Filepath, ZipStoredName = None):
        self.Filepath = Filepath    # Full path to file or ZIP file
        self.ZipStoredName = ZipStoredName # Filename stored in ZIP
        self.songs = []

        # This is false unless the titles file has been locally
        # modified and needs to be flushed.
        self.dirty = False

    def read(self, songDb):
        """ Reads the titles.txt file, and stores the results in the
        indicated db.  This is intended to be called during db
        scan. """

        if self.ZipStoredName != None:
            zip = songDb.GetZipFile(self.Filepath)
            unzipped_data = zip.read(self.ZipStoredName)
            sfile = StringIO(unzipped_data)
            self.__readTitles(songDb, sfile,
                              os.path.join(self.Filepath, self.ZipStoredName))
        else:
            self.__readTitles(songDb, None, self.Filepath)

    def rewrite(self, songDb):
        """ Rewrites the titles.txt file with the current data. """
        if self.ZipStoredName != None:
            sfile = StringIO()
            self.__writeTitles(songDb, sfile,
                               os.path.join(self.Filepath, self.ZipStoredName))
            unzipped_data = sfile.getvalue()
            songDb.DropZipFile(self.Filepath)
            zip = zipfile.ZipFile(self.Filepath, 'a', zipfile.ZIP_DEFLATED)

            # Since the lame Python zipfile.py implementation won't
            # replace an existing file, we have to rename it out of
            # the way.
            self.__renameZipElement(zip, self.ZipStoredName)

            zip.writestr(self.ZipStoredName, unzipped_data)
            zip.close()
        else:
            self.__writeTitles(songDb, None, self.Filepath)

    def __renameZipElement(self, zip, name1, name2 = None):
        """ Renames the file within the archive named "name1" to
        "name2".  To avoid major rewriting of the archive, it is
        required that len(name1) == len(name2).

        If name2 is omitted or None, a new, unique name is
        generated based on the old name.
        """

        zinfo = zip.getinfo(name1)
        zip._writecheck(zinfo)

        if name2 is None:
            # Replace the last letters with digits.
            i = 0
            n = str(i)
            name2 = name1[:-len(n)] + n
            while name2 in zip.NameToInfo:
                i += 1
                n = str(i)
                name2 = name1[:-len(n)] + n

        if len(name1) != len(name2):
            raise RuntimeError("Cannot change length of name with rename().")

        filepos = zip.fp.tell()

        zip.fp.seek(zinfo.header_offset + 30, 0)
        zip.fp.write(name2)
        zinfo.filename = name2

        zip.fp.seek(filepos, 0)

    def __readTitles(self, songDb, catalogFile, catalogPathname):
        self.songs = []
        dirname = os.path.split(catalogPathname)[0]

        if catalogFile == None:
            # Open the file for reading.
            try:
                catalogFile = open(catalogPathname, "rU")
            except:
                print("Could not open titles file %s" % (repr(catalogPathname)))
                return

        for line in catalogFile:
            try:
                line = line.decode('utf-8').strip()
            except UnicodeDecodeError:
                line = line.decode('utf-8', 'replace')
                print("Invalid characters in %s:\n%s" % (repr(catalogPathname), line))

            if line:
                tuple = line.split('\t')
                if len(tuple) == 2:
                    filename, title = tuple
                    artist = ''
                elif len(tuple) == 3:
                    filename, title, artist = tuple
                else:
                    print("Invalid line in %s:\n%s" % (repr(catalogPathname), line))
                    continue

                # Allow a forward slash in the file to stand in for
                # whatever the OS's path separator is.
                filename = filename.replace('/', os.path.sep)

                pathname = os.path.join(dirname, filename)
                song = songDb.filesByFullpath.get(pathname, None)
                if song is None:
                    print("Unknown file in %s:\n%s" % (repr(catalogPathname), repr(filename)))
                else:
                    song.titles = self
                    self.songs.append(song)

                    song.Title = title.strip()
                    song.Artist = artist.strip()
                    if song.Title:
                        songDb.GotTitles = True
                    if song.Artist:
                        songDb.GotArtists = True

    def __makeRelTo(self, filename, relTo):
        """ Returns the filename expressed as a relative path to
        relTo.  Both file paths should be full paths; relTo should
        already have had normcase and normpath applied to it, and
        should end with a slash. """

        filename = os.path.normpath(filename)
        norm = os.path.normcase(filename)
        prefix = os.path.commonprefix((norm, relTo))

        # The common prefix must end with a slash.
        slash = prefix.rfind(os.sep)
        if slash != -1:
            prefix = prefix[:slash + 1]

        filename = filename[len(prefix):]
        relTo = relTo[len(prefix):]

        numSlashes = relTo.count(os.sep)
        if numSlashes > 1:
            backup = '..' + os.sep
            filename = backup * (numSlashes - 1) + filename

        return filename

    def __writeTitles(self, songDb, catalogFile, catalogPathname):
        dirname = os.path.split(catalogPathname)[0]

        if catalogFile == None:
            # Open the file for writing.
            try:
                catalogFile = open(catalogPathname, "w")
            except:
                print("Could not rewrite titles file %s" % (repr(catalogPathname)))
                return

        relTo = os.path.normcase(os.path.normpath(catalogPathname))
        if relTo[-1] != os.sep:
            relTo += os.sep

        for song in self.songs:
            filename = song.Filepath
            if song.ZipStoredName:
                filename = os.path.join(filename, song.ZipStoredName)

            filename = self.__makeRelTo(filename, relTo)

            # Use forward slashes instead of the native separator, to
            # make a more platform-independent titles.txt file.
            filename = filename.replace(os.sep, '/')

            line = filename
            if songDb.GotTitles or songDb.GotArtists:
                line += '\t' + song.Title
            if songDb.GotArtists:
                line += '\t' + song.Artist

            line = line.encode('utf-8')
            catalogFile.write(line + '\n')

class FontData:
    """ This stores the font description selected by the user.
    Hopefully it is enough information to be used both in wx and in
    pygame to reference a unique font on the system. """

    def __init__(self, name = None, size = None, bold = False, italic = False):
        # name may be either a system font name (if size != None) or a
        # filename (if size == None).
        self.name = name
        self.size = size
        self.bold = bold
        self.italic = italic

    def __repr__(self):
        if not self.size:
            return "FontData(%s)" % (repr(self.name))
        else:
            return "FontData(%s, %s, %s, %s)" % (
                repr(self.name), repr(self.size), repr(self.bold), repr(self.italic))

    def getDescription(self):
        desc = self.name
        if self.size:
            desc += ',%spt' % (self.size)
        if self.bold:
            desc += ',bold'
        if self.italic:
            desc += ',italic'

        return desc



# SettingsStruct used as storage only for settings. The instance
# can be pickled to save all user's settings.
class SettingsStruct:

    # This is the list of the encoding strings we offer the user to
    # select from.  You can also type your own.
    Encodings = [
        'cp1252',
        'iso-8859-1',
        'iso-8859-2',
        'iso-8859-5',
        'iso-8859-7',
        'utf-8',
        ]

    # This is the set of CDG zoom modes.
    Zoom = [
        'quick', 'int', 'full', 'soft', 'none',
        ]
    ZoomDesc = {
        'quick' : 'a pixelly scale, maintaining aspect ratio',
        'int' : 'like quick, reducing artifacts a little',
        'full' : 'like quick, but stretches to fill the entire window',
        'soft' : 'a high-quality scale, but may be slow on some hardware',
        'none' : 'keep the display in its original size',
        }

    # Some audio cards seem to support only a limited set of sample
    # rates.  Here are the suggested offerings.
    SampleRates = [
        48000,
        44100,
        22050,
        11025,
        5512,
        ]

    # A list of possible file name deriving combinations.
    # The combination order is stored and used in the parsing algorithm.
    # Track is assumed to be exactly 2 digits.
    FileNameCombinations = [
        'Disc-Track-Artist-Title',
        'DiscTrack-Artist-Title',
        'Disc-Artist-Title',
        'Artist-Title'
        ]

    def __init__(self):
        self.Version = SETTINGS_VERSION

        # Set the default settings, in case none are stored on disk
        self.FolderList = []
        self.CdgExtensions = [ '.cdg' ]
        self.KarExtensions = [ '.kar', '.mid' ]
        self.MpgExtensions = [ '.mpg', '.mpeg', '.avi' ]
        self.IgnoredExtensions = []
        self.LookInsideZips = True
        self.ReadTitlesTxt = True
        self.CheckHashes = False
        self.DeleteIdentical = False
        if env == ENV_WINDOWS:
            self.FilesystemCoding = 'cp1252'
        else:
            self.FilesystemCoding = 'iso-8859-1'
        self.ZipfileCoding = 'cp1252'

        self.WindowSize = (640, 480) # Size of the window for PyKaraoke
        self.FullScreen = False # Determines if the karaoke player should be full screen
        self.NoFrame = False # Determies if the karaoke player should have a window frame.

        # SDL specific parameters; some settings may work better on
        # certain hardware than others
        self.DoubleBuf = True
        self.HardwareSurface = True
        
        self.PlayerSize = (640, 480) # Size of the karaoke player
        self.PlayerPosition = None # Initial position of the karaoke player
        
        self.SplitVertically = True
        self.AutoPlayList = True # Enables or disables the auto play on the play-list
        self.DoubleClickPlayList = True # Enables or disables the double click for playing from the play-list
        self.ClearFromPlayList = True # Enables or disables clearing the playlist with a right click on the play list
        self.Kamikaze = False # Enables or disables the kamikaze button
        self.UsePerformerName = False # Enables or disables the prompting for a performers name.
        self.PlayFromSearchList = True # Enables or disables the playing of a song from the search list
        self.DisplayArtistTitleCols = False # Enables or disables display of artist/title columns

        self.SampleRate = 44100
        self.NumChannels = 2
        self.BufferMs = 50
        self.UseMp3Settings = True

        # This value is a time in milliseconds that will be used to
        # shift the time of the lyrics display relative to the video.
        # It is adjusted by the user pressing the left and right
        # arrows during singing, and is persistent during a session.
        # Positive values make the lyrics anticipate the music,
        # negative values delay them.
        self.SyncDelayMs = 0

        # KAR/MID options
        self.KarEncoding = 'cp1252'  # Default text encoding in karaoke files
        self.KarFont = FontData("DejaVuSans.ttf")
        self.KarBackgroundColour = (0, 0, 0)
        self.KarReadyColour = (255,50,50)
        self.KarSweepColour = (255,255,255)
        self.KarInfoColour = (0, 0, 200)
        self.KarTitleColour = (100, 100, 255)
        self.MIDISampleRate = 44100

        # CDG options
        self.CdgZoom = 'int'
        self.CdgUseC = True
        self.CdgDeriveSongInformation = False # Determines if we should parse file names for song information
        self.CdgFileNameType = -1 # The style index we are using for the file name parsing
        self.ExcludeNonMatchingFilenames = False # Exclude songs from database if can't derive song info

        # MPEG options
        self.MpgNative = True
        self.MpgExternalThreaded = True
        self.MpgExternal = 'mplayer -fs "%(file)s"'

        if env == ENV_WINDOWS:
            self.MpgExternal = '"C:\\Program Files\\Windows Media Player\\wmplayer.exe" "%(file)s" /play /close /fullscreen'
        elif env == ENV_GP2X:
            self.FullScreen = True
            self.PlayerSize = (320, 240)
            self.CdgZoom = 'none'
            # Reduce the default sample rate on the GP2x to save time.
            self.MIDISampleRate = 11025
            self.MpgExternal = './mplayer_cmdline "%(file)s"'
            self.MpgExternalThreaded = False
            self.BufferMs = 250

            # Define the CPU speed for various activities.  We're
            # conservative here and avoid overclocking by default.
            # The user can push these values higher if he knows his
            # GP2X can handle it.
            self.CPUSpeed_startup = 240
            self.CPUSpeed_wait = 33
            self.CPUSpeed_menu_idle = 33
            self.CPUSpeed_menu_slow = 100
            self.CPUSpeed_menu_fast = 240
            self.CPUSpeed_load = 240
            self.CPUSpeed_cdg = 200
            self.CPUSpeed_kar = 240
            self.CPUSpeed_mpg = 200

# This is a trivial class used to wrap the song database with a
# version number.
class DBStruct:
    def __init__(self):
        self.Version = DATABASE_VERSION
        pass

# Song database class with methods for building the database, searching etc
class SongDB:
    def __init__(self):
        # Filepaths and titles are stored in a list of SongStruct instances
        self.FullSongList = []

        # This is the same list, with songs of the same artist/title
        # removed.
        self.UniqueSongList = []

        # Here's those lists again, cached into various different
        # sorts.
        self.SortedLists = {}

        # And this is just the currently-active song list, according
        # to selected sort.
        self.SongList = []

        # The list of TitlesFiles we have found in our scan.
        self.TitlesFiles = []

        # A cache of zip files.
        self.ZipFiles = []

        # Set true if there are local changes to the database that
        # need to be saved to disk.
        self.databaseDirty = False

        # Some databases may omit either or both of titles and
        # artists, relying on filenames instead.
        self.GotTitles = False
        self.GotArtists = False

        # Create a SettingsStruct instance for storing settings
        # in case none are stored.
        self.Settings = SettingsStruct()

        # All temporary files use this prefix
        self.TempFilePrefix = "00Pykar__"

        self.SaveDir = self.getSaveDirectory()
        self.TempDir = self.getTempDirectory()
        self.CleanupTempFiles()

    def getSaveDirectory(self):
        """ Returns the directory in which the settings files should
        be saved. """

        # If we have PYKARAOKE_DIR defined, use it.
        dir = os.getenv('PYKARAOKE_DIR')
        if dir:
            return dir

        if env == ENV_GP2X:
            # On the GP2X, just save db files in the root directory.
            # Makes it easier to find them, and avoids directory
            # clutter.
            return '.'

        # Without PYKARAOKE_DIR, use ~/.pykaraoke.  Try to figure that
        # out.
        homeDir = self.getHomeDirectory()
        return os.path.join(homeDir, ".pykaraoke")

    def getTempDirectory(self):
        """ Returns the directory in which temporary files should be
        saved. """
        dir = os.getenv('PYKARAOKE_TEMP_DIR')
        if dir:
            return dir

        dir = os.getenv('TEMP')
        if dir:
            return os.path.join(dir, 'pykaraoke')

        if env != ENV_WINDOWS:
            if os.path.exists('/tmp'):
                return '/tmp/pykaraoke'
        else:
            try:
                import win32api
                return os.path.join(win32api.GetTempPath(), 'pykaraoke')
            except:
                pass

        # If we can't find a good temp directory, use our save directory.
        return self.getSaveDirectory()

    def getHomeDirectory(self):
        """ Returns the user's home directory, if we can figure that
        out. """

        if env != ENV_GP2X:
            # First attempt: ask wx, if it's available.
            try:
                import wx
                return wx.GetHomeDir()
            except:
                pass

            # Second attempt: look in $HOME
            home = os.getenv('HOME')
            if home:
                return home

        # Give up and return the current directory.
        return '.'

    def makeSongStruct(self, filename):
        """ Creates a quick SongStruct representing the indicated
        filename.  The file may be embedded within a zip file; treat
        the zip filename as a directory in this case. """

        # Is this a file within a zip file?
        zipStoredName = None
        z = filename.find('.zip/')
        if z == -1:
            z = filename.find('.zip' + os.path.sep)
        if z != -1:
            zipStoredName = filename[z + 5:]
            filename = filename[:z + 4]

        song = SongStruct(filename, self.Settings,
                          ZipStoredName = zipStoredName)
        return song


    def chooseTitles(self, song):
        """ Chooses an appropriate titles file to represent the
        indicated song file.  If there is no appropriate titles file,
        creates one.  Applies the song to the new titles file. """

        if song.titles:
            # This song already has a titles file.
            return

        songPath = song.Filepath
        if song.ZipStoredName:
            songPath = os.path.join(songPath, song.ZipStoredName)

        relTo = os.path.normcase(os.path.normpath(songPath))

        # Look for the titles file, in a directory above the song,
        # with the longest prefix in common with the song.  This will
        # be the best titles file.
        bestTitles = None
        bestPrefix = ''
        for titles in self.TitlesFiles:
            titlesPath = titles.Filepath
            if titles.ZipStoredName:
                titlesPath = os.path.join(titlesPath, titles.ZipStoredName)
            norm = os.path.normcase(os.path.normpath(titlesPath))

            prefix = os.path.commonprefix((norm, relTo))
            # The common prefix must end with a slash.
            slash = prefix.rfind(os.sep)
            if slash != -1:
                prefix = prefix[:slash + 1]

            norm = norm[len(prefix):]
            if os.path.sep in norm:
                # This titles file is in a subordinate directory.
                # Skip it.
                continue

            if len(prefix) > len(bestPrefix):
                bestTitles = titles
                bestPrefix = prefix

        if not bestTitles:
            # Didn't find a good candidate.  Create a new titles file,
            # in the root of whichever directory contains the song.
            bestDir = None
            for dir in self.Settings.FolderList:
                norm = os.path.normcase(os.path.normpath(dir))
                if relTo.startswith(norm):
                    bestDir = dir
                    break

            if not bestDir:
                # No folder!  Put it with the song itself.
                bestDir = os.path.splitext(song.Filepath)[0]

            bestTitles = TitleStruct(os.path.join(bestDir, 'titles.txt'))
            self.TitlesFiles.append(bestTitles)

        bestTitles.songs.append(song)
        song.titles = bestTitles
        bestTitles.dirty = True
        self.databaseDirty = True

    def LoadSettings (self, errorCallback):
        """ Load the personal settings (but not yet the database). """

        # Load the settings file
        settings_filepath = os.path.join (self.SaveDir, "settings.dat")
        if os.path.exists (settings_filepath):
            file = open(settings_filepath, "rU")
            loadsettings = SettingsStruct()
            # We use eval to evaluate the settings file.  This is
            # very easy and powerful, though it does mean we
            # execute whatever arbitrary Python code a malicious
            # user might have put in there.  On the other hand,
            # it's the user's own machine, and this isn't any less
            # secure than using cPickle to decode the file (which
            # basically does the same thing anyway).
            for line in file:
                if '=' not in line:
                    continue
                key, value = line.split('=', 1)
                key = key.strip()

                # Ignore any definitions for keys we don't already
                # have.  This allows us to phase out old config values
                # we no longer use.
                if not hasattr(loadsettings, key):
                    continue

                try:
                    value = eval(value)
                except:
                    # Ignore anything that isn't valid Python.
                    print("Invalid value for %s" % (key))
                    continue

                setattr(loadsettings, key, value)

            # Check settings are for the current version.
            message = None
            if loadsettings:
                if loadsettings.Version == SETTINGS_VERSION:
                    self.Settings = loadsettings
                else:
                    message = "New version of PyKaraoke, clearing settings"

            if message:
                if errorCallback:
                    errorCallback(message)
                else:
                    print(message)

    def LoadDatabase(self, errorCallback):
        """ Load the saved database. """

        self.FullSongList = []
        self.UniqueSongList = []
        self.TitlesFiles = []
        self.GotTitles = False
        self.GotArtists = False

        # Load the database file
        db_filepath = os.path.join (self.SaveDir, "songdb.dat")
        if os.path.exists (db_filepath):
            file = open (db_filepath, "rb")
            loaddb = None
            try:
                loaddb = pickle.load (file)
            except:
                pass
            if (getattr(loaddb, 'Version', None) == DATABASE_VERSION):
                self.FullSongList = loaddb.FullSongList
                self.SongList = loaddb.FullSongList
                self.UniqueSongList = loaddb.UniqueSongList
                self.TitlesFiles = loaddb.TitlesFiles
                self.GotTitles = loaddb.GotTitles
                self.GotArtists = loaddb.GotArtists
            else:
                if errorCallback:
                   errorCallback("New version of PyKaraoke, clearing database")

        self.databaseDirty = False

##         # This forces the titles files to be rewritten at the next
##         # "save" operation.
##         for titles in self.TitlesFiles:
##             titles.dirty = True
##         self.databaseDirty = True

    def SaveSettings (self):
        """ Save user settings to the home directory. """

        # Create the temp directory if it doesn't exist already
        if not os.path.exists (self.SaveDir):
            os.mkdir(self.SaveDir)

        # Save the settings file
        settings_filepath = os.path.join (self.SaveDir, "settings.dat")
        try:
            file = open (settings_filepath, "w")
        except IOError as message:
            print(message)
        else:
            # We don't use pickle to dump out the settings anymore.
            # Instead, we write them in this human-readable and
            # human-editable format.
            keys = list(self.Settings.__dict__.keys())
            keys.sort()
            for k in keys:
                if not k.startswith('__'):
                    value = getattr(self.Settings, k)
                    print("%s = %s" % (k, repr(value)), file=file)

    def SaveDatabase(self):
        """ Save the database to the appropriate directory. """

        if not self.databaseDirty:
            return

        try:
            # Create the temp directory if it doesn't exist already
            if not os.path.exists (self.SaveDir):
                os.mkdir(self.SaveDir)

            # Write out any titles files that have changed.
            for titles in self.TitlesFiles:
                if titles.dirty:
                    titles.rewrite(self)
                    titles.dirty = False

            # Check for newly unique files
            self.makeUniqueSongs()

            # Save the database file
            db_filepath = os.path.join (self.SaveDir, "songdb.dat")
            file = open (db_filepath, "wb")

            loaddb = DBStruct()
            loaddb.FullSongList = self.FullSongList
            loaddb.UniqueSongList = self.UniqueSongList
            loaddb.TitlesFiles = self.TitlesFiles
            loaddb.GotTitles = self.GotTitles
            loaddb.GotArtists = self.GotArtists

            pickle.dump (loaddb, file, pickle.HIGHEST_PROTOCOL)
        except IOError as message:
            print(message)
        self.databaseDirty = False

    def GetSong(self, index):
        """ This returns the song stored in index in the database. """
        return self.FullSongList[index]

    def BuildSearchDatabase(self, yielder, busyDlg):
        # Zap the database and build again from scratch. Return True
        # if was cancelled.
        self.FullSongList = []
        self.TitlesFiles = []

        return self.doSearch(self.Settings.FolderList, yielder, busyDlg)

    def AddFile(self, filename):
        """Adds just the indicated file to the DB.  If the file is a
        directory or a zip file, recursively scans within it and adds
        all the sub-files. """

        self.doSearch([filename], AppYielder(), BusyCancelDialog())

    def GetZipFile(self, filename):
        """Creates a ZipFile object corresponding to the indicated zip
        file on disk and returns it.  If there was already a ZipFile
        for this filename in the cache, just returns that one
        instead, saving on load time from repeatedly loading the same
        zip file. """

        for tuple in self.ZipFiles:
            cacheFilename, cacheZip = tuple
            if cacheFilename == filename:
                # Here is a zip file in the cache; move it to the
                # front of the list.
                self.ZipFiles.remove(tuple)
                self.ZipFiles.insert(0, tuple)
                return cacheZip

        # The zip file was not in the cache, create a new one and cache it.
        zip = zipfile.ZipFile(filename)
        if len(self.ZipFiles) >= MAX_ZIP_FILES:
            del self.ZipFiles[-1]
        self.ZipFiles.insert(0, (filename, zip))
        return zip

    def DropZipFile(self, filename):
        """Releases an opened zip file by the indicated filename, if any. """

        for tuple in self.ZipFiles:
            cacheFilename, cacheZip = tuple
            if cacheFilename == filename:
                # Here is the zip file in the cache; remove it.
                self.ZipFiles.remove(tuple)
                return

    def doSearch(self, fileList, yielder, busyDlg):
        """ This is the actual implementation of BuildSearchDatabase()
        and AddFile()."""

        self.BusyDlg = busyDlg
        self.BusyDlg.SetProgress("Scanning", 0.0)
        yielder.Yield()
        self.BusyDlg.Show()

        self.lastBusyUpdate = time.time()
        self.filesByFullpath = {}

        for i in range(len(fileList)):
            root_path = fileList[i]

            # Assemble a stack of progress amounts through the various
            # directory levels.  This way we can update a progress bar
            # without knowing exactly how many directories we are
            # going to traverse.  We give each directory equal weight
            # regardless of the number of files within it.
            progress = [(i, len(fileList))]
            self.fileScan(root_path, progress, yielder)
            if self.BusyDlg.Clicked:
                break

        if self.TitlesFiles and not self.BusyDlg.Clicked:
            self.BusyDlg.SetProgress("Reading titles files", 0.0)
            yielder.Yield()
            self.lastBusyUpdate = time.time()

            # Now go back and read any titles.txt files we came across.
            # These will have meta-information about the files, such as
            # the title and/or artist.
            for i in range(len(self.TitlesFiles)):
                if self.BusyDlg.Clicked:
                    break
                now = time.time()
                if now - self.lastBusyUpdate > 0.1:
                    # Every so often, update the current path on the display.
                    self.BusyDlg.SetProgress(
                        "Reading titles files",
                        float(i) / float(len(self.TitlesFiles)))
                    yielder.Yield()
                    self.lastBusyUpdate = now
                self.TitlesFiles[i].read(self)

        if self.Settings.CheckHashes:
            self.checkFileHashes(yielder)

        self.BusyDlg.SetProgress("Finalizing", 1.0)
        yielder.Yield()

        # This structure was just temporary, for use just while
        # scanning the directories.  Remove it now.
        del self.filesByFullpath

        self.makeUniqueSongs()
        self.databaseDirty = True

        cancelled = self.BusyDlg.Clicked
        self.BusyDlg.Destroy()

        return cancelled

    def folderScan (self, FolderToScan, progress, yielder):
        # Search for karaoke files inside the folder, looking inside ZIPs if
        # configured to do so. Function is recursive for subfolders.
        try:
            filedir_list = os.listdir(FolderToScan)
        except:
            print("Couldn't scan %s" % (repr(FolderToScan)))
            return False

        # Sort the list, using printable strings for the sort key to
        # prevent issues with unicode characters in non-unicode strings
        # in the list
        filedir_list.sort(key=repr)

        # Loop through the list
        for i in range(len(filedir_list)):
            item = filedir_list[i]
            if self.BusyDlg.Clicked:
                return True

            # Allow windows to refresh now and again while scanning
            yielder.ConsiderYield()

            # Build the full file path. Check file types match, as
            # os.listdir() can return non-unicode while the folder
            # is still unicode.
            if (type(FolderToScan) != type(item)):
                full_path = os.path.join(str(FolderToScan), str(item))
                print("Folder %s and file %s do not match types" % (repr(FolderToScan), repr(item)))
            else:
                full_path = os.path.join(FolderToScan, item)

            nextProgress = progress + [(i, len(filedir_list))]
            self.fileScan(full_path, nextProgress, yielder)
            if self.BusyDlg.Clicked:
                return

    def __computeProgressValue(self, progress):
        """ Returns a floating-point value in the range 0 to 1 that
        corresponds to the progress list we have built up while
        traversing the directory structure hierarchically.  This is
        used to update the progress bar linearly while we traverse the
        hierarchy. """

        # The progress list is a list of tuples of the form [(i0,
        # len0), (i1, len1), (i2, len2), ..., (in, lenn)].  There is
        # one entry for each directory level we have visited.

        # We need to boil this down into a single nondecreasing
        # number.  A simple mathematical series.

        range = 1.0
        result = 0.0
        for i, len in progress:
            if len > 1:
                result += range * (float(i) / float(len))
                range = range * (1.0 / float(len))
        return result

    def fileScan(self, full_path, progress, yielder):
        now = time.time()
        if now - self.lastBusyUpdate > 0.1:
            # Every so often, update the progress bar.
            basename = os.path.split(full_path)[1]
            # Sanitise byte-strings
            try:
                basename = str(basename)
            except UnicodeDecodeError:
                basename = basename.decode("ascii", "replace")
            self.BusyDlg.SetProgress(
                "Scanning %s" % basename,
                self.__computeProgressValue(progress))
            yielder.Yield()
            self.lastBusyUpdate = now

        # Recurse into subdirectories
        if os.path.isdir(full_path):
            basename = os.path.split(full_path)[1]
            if basename == 'CVS' or basename == '.svn':
                # But skip over these bogus directories.
                pass
            else:
                self.folderScan(full_path, progress, yielder)
            if self.BusyDlg.Clicked:
                return
        # Store file details if it's a file type we're interested in
        else:
            root, ext = os.path.splitext(full_path)
            # Non-ZIP files
            if self.Settings.ReadTitlesTxt and full_path.endswith('titles.txt'):
                # Save this titles.txt file for reading later.
                self.TitlesFiles.append(TitleStruct(full_path))
            elif self.IsExtensionValid(ext):
                try:
                    self.addSong(SongStruct(full_path, self.Settings, DatabaseAdd = True))
                except KeyError:
                    print("Excluding filename with unexpected format: %s " % repr(os.path.basename(full_path)))
            # Look inside ZIPs if configured to do so
            elif self.Settings.LookInsideZips and ext.lower() == ".zip":
                try:
                    if zipfile.is_zipfile(full_path):
                        zip = self.GetZipFile(full_path)
                        namelist = zip.namelist()
                        for i in range(len(namelist)):
                            filename = namelist[i]

                            now = time.time()
                            if now - self.lastBusyUpdate > 0.1:
                                # Every so often, update the progress bar.
                                nextProgress = progress + [(i, len(namelist))]
                                basename = os.path.split(full_path)[1]
                                # Sanitise byte-strings
                                try:
                                    basename = str(basename)
                                except UnicodeDecodeError:
                                    basename = basename.decode("ascii", "replace")
                                self.BusyDlg.SetProgress(
                                    "Scanning %s" % basename,
                                    self.__computeProgressValue(nextProgress))
                                yielder.Yield()
                                self.lastBusyUpdate = now

                            root, ext = os.path.splitext(filename)
                            if self.Settings.ReadTitlesTxt and filename.endswith('titles.txt'):
                                # Save this titles.txt file for reading later.
                                self.TitlesFiles.append(TitleStruct(full_path, ZipStoredName = filename))
                            elif self.IsExtensionValid(ext):
                                # Python zipfile only supports deflated and stored
                                info = zip.getinfo(filename)
                                if info.compress_type == zipfile.ZIP_STORED or info.compress_type == zipfile.ZIP_DEFLATED:
                                    #print ("Adding song %s in ZIP file %s"%(repr(filename), repr(full_path)))
                                    try:
                                        self.addSong(SongStruct(full_path, self.Settings, ZipStoredName = filename, DatabaseAdd = True))
                                    except KeyError:
                                        print("Excluding filename with unexpected format: %s " % repr(os.path.basename(full_path)))
                                else:
                                    print(("ZIP member compressed with unsupported type (%d): %s"%(info.compress_type, repr(full_path))))
                    else:
                        print("Cannot parse ZIP file: " + repr(full_path))
                except:
                    print("Error looking inside zip " + repr(full_path))

    # Add a folder to the database search list
    def FolderAdd (self, FolderPath):
        if FolderPath not in self.Settings.FolderList:
            self.Settings.FolderList.append(FolderPath)

    # Remove a folder from the database search list
    def FolderDel (self, FolderPath):
        self.Settings.FolderList.remove(FolderPath)

    # Get the list of folders currently set up for the database
    def GetFolderList (self):
        return self.Settings.FolderList

    # Search the database for occurrences of the search terms.
    # If there are multiple terms, all must exist for a match.
    # The search is case-insensitive and searches both the title
    # and the pathname.
    # Returns a list of SongStruct instances.
    def SearchDatabase (self, SearchTerms, yielder):
        # Display a busy cursor while searching, yielding now and again
        # to update the GUI.
        ResultsList = []
        LowerTerms = SearchTerms.lower()
        TermsList = LowerTerms.split()
        for song in self.FullSongList:
            yielder.ConsiderYield()
            LowerTitle = song.Title.lower()
            LowerArtist = song.Artist.lower()
            LowerPath = song.DisplayFilename.lower()
            # If it's a zip file, also include the zip filename
            if song.ZipStoredName:
                LowerZipName = os.path.basename(song.Filepath).lower()
            else:
                LowerZipName = ""
            misses = 0
            for term in TermsList:
                try:
                    if (term not in LowerTitle) and \
                       (term not in LowerArtist) and \
                       (term not in LowerZipName) and \
                       (term not in LowerPath):
                        misses = misses + 1
                except UnicodeDecodeError:
                    print("Unicode error looking up %s in %s" % (repr(term), repr(LowerZipName)))
                    misses = misses + 1 
            if misses == 0:
                ResultsList.append(song)
        return ResultsList

    # Get the song database size (number of songs)
    def GetDatabaseSize (self):
        return len(self.FullSongList)

    # Check if the passed file extension is configured for this database
    def IsExtensionValid (self, extension):
        ext = extension.lower()
        if ext in self.Settings.IgnoredExtensions:
            return False
        if ext in self.Settings.KarExtensions or \
           ext in self.Settings.CdgExtensions or \
           ext in self.Settings.MpgExtensions:
            return True
        return False

    # Create a directory for use by PyKaraoke for temporary zip files
    # and for saving the song database and settings.
    # This will be under the Wx idea of the home directory.
    def CreateTempDir (self):
        if not os.path.exists(self.TempDir):
            os.mkdir (self.TempDir)

    # Create temporary filename prefix. Returns a path and filename
    # prefix which can be used as a base string for temporary files.
    # You must clean them up when done using CleanupTempFiles().
    # Also automatically creates a temporary directory if one doesn't
    # exist already.
    def CreateTempFileNamePrefix (self):
        self.CreateTempDir()
        full_prefix = os.path.join (self.TempDir, self.TempFilePrefix)
        return full_prefix

    # Clean up any temporary (unzipped) files on startup/exit/end of song
    def CleanupTempFiles (self):
        if os.path.exists (self.TempDir):
            filedir_list = os.listdir(self.TempDir)
            for item in filedir_list:
                if item.startswith(self.TempFilePrefix):
                    full_path = os.path.join (self.TempDir, item)
                    try:
                        os.unlink(full_path)
                    except:
                        # The unlink can fail on Windows due to a bug in
                        # pygame.mixer.music which does not release the
                        # file handle until you load another music file.
                        pass

    def SelectSort(self, sort, allowResort = True):
        """Sorts the list of songs in order according to the indicated
        key, which must be one of 'title', 'artist', or 'filename'.
        Also sets self.GetSongTuple to a functor which, when called, returns a
        3-tuple of strings suitable for displaying for each song,
        where the first string of each tuple is the sort key.

        This may require re-sorting the list on-the-fly.  If
        allowResort is False, then the list will never be re-sorted;
        rather, the method will return True if the sort was
        successfully applied, or False if a re-sort was necessary but
        not performed, in which case the list retains its original
        sort. """

        if sort == 'title' and self.GotTitles:
            if self.GotArtists:
                getSongTuple = self.getSongTupleTitleArtistFilename
                sortKeys = ('title', 'artist', 'filename')
                getSortKey = self.getSongTupleTitleArtistFilenameSortKey
            else:
                getSongTuple = self.getSongTupleTitleFilenameArtist
                sortKeys = ('title', 'filename')
                getSortKey = self.getSongTupleTitleFilenameArtistSortKey

        elif sort == 'artist' and self.GotArtists:
            if self.GotTitles:
                getSongTuple = self.getSongTupleArtistTitleFilename
                sortKeys = ('artist', 'title', 'filename')
                getSortKey = self.getSongTupleArtistTitleFilenameSortKey
            else:
                getSongTuple = self.getSongTupleArtistFilenameTitle
                sortKeys = ('artist', 'filename')
                getSortKey = self.getSongTupleArtistFilenameTitleSortKey

        else: # filename
            sort = 'filename'
            if self.GotTitles and self.GotArtists:
                getSongTuple = self.getSongTupleFilenameTitleArtist
                sortKeys = ('filename', 'title', 'artist')
                getSortKey = self.getSongTupleFilenameTitleArtistSortKey
            elif self.GotTitles:
                getSongTuple = self.getSongTupleFilenameTitleArtist
                sortKeys = ('filename', 'title')
                getSortKey = self.getSongTupleFilenameTitleArtistSortKey
            elif self.GotArtists:
                getSongTuple = self.getSongTupleFilenameArtistTitle
                sortKeys = ('filename', 'artist')
                getSortKey = self.getSongTupleFilenameArtistTitleSortKey
            else:
                getSongTuple = self.getSongTupleFilenameArtistTitle
                sortKeys = ('filename',)
                getSortKey = self.getSongTupleFilenameArtistTitleSortKey

        list = self.SortedLists.get(getSortKey, None)
        if list is None:
            if not allowResort:
                # Return False to indicate that a sort was not applied.
                return False
            
            # We haven't asked for this sort key before; we have to
            # sort the list now.  Once sorted, we can keep it around
            # for future requests.
            if sort == 'filename':
                list = self.FullSongList[:]
            else:
                list = self.UniqueSongList[:]
            list.sort(key = getSortKey)
            self.SortedLists[getSortKey] = list

        global fileSortKey
        fileSortKey = getSortKey

        self.Sort = sort
        self.SortKeys = sortKeys
        self.GetSongTuple = getSongTuple
        self.GetSortKey = getSortKey

        self.SongList = list

        return True

    def getSongTupleTitleArtistFilename(self, file):
        return (file.Title, file.Artist, file.getDisplayFilenames())

    def getSongTupleTitleFilenameArtist(self, file):
        return (file.Title, file.getDisplayFilenames(), file.Artist)

    def getSongTupleArtistTitleFilename(self, file):
        return (file.Artist, file.Title, file.getDisplayFilenames())

    def getSongTupleArtistFilenameTitle(self, file):
        return (file.Artist, file.getDisplayFilenames(), file.Title)

    def getSongTupleFilenameTitleArtist(self, file):
        return (file.DisplayFilename, file.Title, file.Artist)

    def getSongTupleFilenameArtistTitle(self, file):
        return (file.DisplayFilename, file.Artist, file.Title)

    def getSongTupleTitleArtistFilenameSortKey(self, file):
        return (file.MakeSortKey(file.Title), file.MakeSortKey(file.Artist), file.getDisplayFilenames().lower(), id(file))

    def getSongTupleTitleFilenameArtistSortKey(self, file):
        return (file.MakeSortKey(file.Title), file.DisplayFilename.lower(), file.MakeSortKey(file.Artist), id(file))

    def getSongTupleArtistTitleFilenameSortKey(self, file):
        return (file.MakeSortKey(file.Artist), file.MakeSortKey(file.Title), file.DisplayFilename.lower(), id(file))

    def getSongTupleArtistFilenameTitleSortKey(self, file):
        return (file.MakeSortKey(file.Artist), file.DisplayFilename.lower(), file.MakeSortKey(file.Title), id(file))

    def getSongTupleFilenameTitleArtistSortKey(self, file):
        return (file.DisplayFilename.lower(), file.MakeSortKey(file.Title), file.MakeSortKey(file.Artist), id(file))

    def getSongTupleFilenameArtistTitleSortKey(self, file):
        return (file.DisplayFilename.lower(), file.MakeSortKey(file.Artist), file.MakeSortKey(file.Title), id(file))

    def addSong(self, file):
        self.FullSongList.append(file)
        if file.Title:
            self.GotTitles = True
        if file.Artist:
            self.GotArtists = True

        if hasattr(self, 'filesByFullpath'):
            # Also record the file in the temporary map by fullpath name,
            # so we can cross-reference it with a titles.txt file that
            # might reference it.
            fullpath = file.Filepath
            if file.ZipStoredName:
                name = file.ZipStoredName.replace('/', os.path.sep)
                fullpath = os.path.join(fullpath, name)
            self.filesByFullpath[fullpath] = file

    def checkFileHashes(self, yielder):
        """ Walks through self.FullSongList, checking for md5 hashes
        to see if any files are duplicated. """

        self.BusyDlg.SetProgress("Checking file hashes", 0.0)
        yielder.Yield()
        self.lastBusyUpdate = time.time()
        numDuplicates = 0

        # Check the md5's of each file, to see if there are any
        # duplicates.
        fileHashes = {}
        numFiles = len(self.FullSongList)
        for i in range(numFiles):
            now = time.time()
            if now - self.lastBusyUpdate > 0.1:
                # Every so often, update the progress bar.
                label = "Checking file hashes"
                if numDuplicates:
                    label = "%s duplicates found" % (numDuplicates)
                self.BusyDlg.SetProgress(
                    label, float(i) / float(numFiles))
                yielder.Yield()
                self.lastBusyUpdate = now

            if self.BusyDlg.Clicked:
                return


            # Calculate the MD5 hash of the songfile.
            m = md5()

            # Get details of the associated files
            song = self.FullSongList[i]
            datas = song.GetSongDatas()
            if len(datas) > 0:
                song_data = datas[0]
                # If the data has already been read in, use it directly.
                # Otherwise read the file off disk for temporary use.
                if song_data.data != None:
                    m.update(song_data.data)
                else:
                    f = open(song_data.filename)
                    if f != None:
                        while True:
                            data = f.read(64*1024)
                            if not data:
                                break
                            m.update(data)
                list = fileHashes.setdefault(m.digest(), [])
                if list:
                    numDuplicates += 1
                list.append(i)

        # Remove the identical files from the database.  If specified,
        # remove them from disk too.
        removeIndexes = {}
        for list in list(fileHashes.values()):
            if len(list) > 1:
                filenames = [self.FullSongList[i].DisplayFilename for i in list]
                print("Identical songs: %s" % repr((', '.join(filenames))))
                for i in list[1:]:
                    extra = self.FullSongList[i]
                    removeIndexes[i] = True
                    if extra.titles:
                        extra.titles.dirty = True
                    if self.Settings.DeleteIdentical:
                        if extra.ZipStoredName:
                            # Can't delete a song within a zip, sorry.
                            pass
                        else:
                            os.remove(extra.Filepath)

        # Now rebuild the FullSongList without the removed files.
        newSongList = []
        for i in range(numFiles):
            if i not in removeIndexes:
                newSongList.append(self.FullSongList[i])
        self.FullSongList = newSongList

    def makeUniqueSongs(self):
        """ Walks through self.FullSongList, and builds up
        self.UniqueSongList, which collects only those songs who have
        the same artist/title combination. """

        if not self.GotArtists and not self.GotTitles:
            # A special case: titles.txt files aren't in use.
            self.UniqueSongList = self.FullSongList
            return

        songsByArtistTitle = {}

        for song in self.FullSongList:
            tuple = (song.Artist.lower(), song.Title.lower())
            songList = songsByArtistTitle.setdefault(tuple, [])
            songList.append(song)

        # Now go through and sort each songList into order by type.

        self.UniqueSongList = []
        for songList in list(songsByArtistTitle.values()):
            songList.sort(key = SongStruct.getTypeSort)
            for song in songList:
                song.sameSongs = songList
            self.UniqueSongList.append(songList[0])

globalSongDB = SongDB()
